// -------------------------------------------------------------------------
//
// Copyright or © or Copr. CNRS, (November 16, 2004)
//
// This software is a library providing tool for bioinformatics.
//
// This software is governed by the CeCILL  license under French law and
// abiding by the rules of distribution of free software.  You can  use,
// modify and/ or redistribute the software under the terms of the CeCILL
// license as circulated by CEA, CNRS and INRIA at the following URL
// "http://www.cecill.info".
//
// As a counterpart to the access to the source code and  rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty  and the software's author,  the holder of the
// economic rights,  and the successive licensors  have only  limited
// liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading,  using,  modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean  that it is complicated to manipulate,  and  that  also
// therefore means  that it is reserved for developers  and  experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and,  more generally, to use and operate it in the
// same conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.
//
// -------------------------------------------------------------------------

/**
 * <p>Description : Baobab core classes.</p>
 * <p>Copyright : Copyright (c) Julien Dutheil</p>
 * @author Julien Dutheil
 */

package TreeEditor;

import gnu.jtools.utils.gui.GraphicsTools;
import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.table.*;
import java.awt.event.*;
import java.util.*;

public class ColorManager extends JDialog {

  public ColorManager(Frame parent) {
    super(parent, true);
    this.parent = parent;

    setLocale(parent.getLocale());
    setTitle(Baobab.message.getString("ColorManagerTitle"));

    methods = new JTabbedPane(JTabbedPane.TOP);
    methods.add(Baobab.message.getString("ColorManagerMethod_STDS"), new STDSMethodPanel(this));
    methods.add(Baobab.message.getString("ColorManagerMethod_STNV"), new STNVMethodPanel(this));

    getContentPane().add(methods);
    pack();
  }

  public ColoringMethod getColoringMethod() {
    ColoringMethodPanel cmp = (ColoringMethodPanel)methods.getSelectedComponent();
    return cmp.getColoringMethod();
  }

  class ColorRenderer extends JLabel implements TableCellRenderer {
     Border unselectedBorder = null;
     Border selectedBorder = null;
     boolean isBordered = true;

     public ColorRenderer(boolean isBordered) {
       super();
       this.isBordered = isBordered;
       setOpaque(true); //MUST do this for background to show up.
     }

     public Component getTableCellRendererComponent(
                                JTable table, Object color,
                                boolean isSelected, boolean hasFocus,
                                int row, int column) {
       setBackground((Color)color);
       if (isBordered) {
         if (isSelected) {
           if (selectedBorder == null) {
             selectedBorder = BorderFactory.createMatteBorder(2, 5, 2, 5, table.getSelectionBackground());
           }
           setBorder(selectedBorder);
          } else {
            if (unselectedBorder == null) {
              unselectedBorder = BorderFactory.createMatteBorder(2, 5, 2, 5, table.getBackground());
            }
          setBorder(unselectedBorder);
        }
      }
    return this;
    }
  }

  private void setUpColorRenderer(JTable table) {
    table.setDefaultRenderer(Color.class, new ColorRenderer(true));
  }

  public static class ColorEditor extends DefaultCellEditor {
    Color currentColor = null;

    public ColorEditor(JButton b) {
      super(new JCheckBox()); //Unfortunately, the constructor
                              //expects a check box, combo box,
                              //or text field.
      editorComponent = b;
      setClickCountToStart(1); //This is usually 1 or 2.

      //Must do this so that editing stops when appropriate.
       b.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
           fireEditingStopped();
         }
       });
    }

    protected void fireEditingStopped() {
      super.fireEditingStopped();
    }

    public Object getCellEditorValue() {
      return currentColor;
    }

    public Component getTableCellEditorComponent(JTable table,
                                                 Object value,
                                                 boolean isSelected,
                                                 int row,
                                                 int column) {
      ((JButton)editorComponent).setText(value.toString());
      currentColor = (Color)value;
      return editorComponent;
    }
  }

  //Set up the editor for the Color cells.
  private void setUpColorEditor(JTable table) {
    //First, set up the button that brings up the dialog.
     final JButton button = new JButton("") {
        public void setText(String s) {
          //Button never shows text -- only color.
        }
     };
     button.setBackground(Color.white);
     button.setBorderPainted(false);
     button.setMargin(new Insets(0,0,0,0));

     //Now create an editor to encapsulate the button, and
     //set it up as the editor for all Color cells.
     final ColorEditor colorEditor = new ColorEditor(button);
     table.setDefaultEditor(Color.class, colorEditor);

     //Set up the dialog that the button brings up.
     final JColorChooser colorChooser = new JColorChooser();
     ActionListener okListener = new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          colorEditor.currentColor = colorChooser.getColor();
        }
      };
      final JDialog dialog = JColorChooser.createDialog(button,
                                      Baobab.message.getString("ColorManagerChooserTitle"),
                                      true,
                                      colorChooser,
                                      okListener,
                                      null); //XXXDoublecheck this is OK

      //Here's the code that brings up the dialog.
      button.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          button.setBackground(colorEditor.currentColor);
          colorChooser.setColor(colorEditor.currentColor);
          //Without the following line, the dialog comes up
          //in the middle of the screen.
          //dialog.setLocationRelativeTo(button);
          dialog.show();
          }
      });
  }

  public abstract static class AttributeValuesTableModel extends AbstractTableModel {

    public int getColumnCount() { return columnNames.length; }

    public int getRowCount() { return data[0].size(); }

    public String getColumnName(int col) { return columnNames[col]; }

    public Object getValueAt(int row, int col) { return data[col].get(row); }

    public Class getColumnClass(int c) { return getValueAt(0, c).getClass(); }

    public boolean isCellEditable(int row, int col) { return true; }

    public void setValueAt(Object value, int row, int col) {
      data[col].set(row, value);
      fireTableCellUpdated(row, col);
    }

    public void addRow() {
      data[0].add("new value");
      data[1].add(Color.black);
      fireTableDataChanged();
    }

    public void addRow(String value, Color color) {
      data[0].add(value);
      data[1].add(color);
      fireTableDataChanged();
    }

    public void removeRow(int i) {
      data[0].remove(i);
      data[1].remove(i);
      fireTableDataChanged();
    }

    protected java.util.List[] data;
    protected String[] columnNames;
  }


  /**
   * Available panels:
   */

  public abstract class ColoringMethodPanel extends JPanel {

    public ColoringMethodPanel(Window owner) {
      this.owner = owner;
    }

    abstract ColoringMethod getColoringMethod();

    protected Window owner;
  }






  public class STDSMethodPanel extends ColoringMethodPanel implements ActionListener {

    /**
     * STDS stands for:
     * Single Type with Descriptive State.
     */

    public STDSMethodPanel(Window owner) {
      super(owner);
      type = new JTextField(20);
      JToolBar barTop = new JToolBar(JToolBar.HORIZONTAL);
               barTop.add(new JLabel(Baobab.message.getString("ColorManagerAttribute")));
               barTop.add(type);
               barTop.setFloatable(false);

      ctm = new STDSTableModel();
      table = new JTable(ctm);
      setUpColorEditor(table);
      setUpColorRenderer(table);

      add = new JButton(Baobab.message.getString("ColorManagerAdd"));
      add.addActionListener(this);
      remove = new JButton(Baobab.message.getString("ColorManagerRemove"));
      remove.addActionListener(this);
      ok = new JButton(Baobab.message.getString("ColorManagerApply"));
      ok.addActionListener(this);
      toolBar = new JToolBar(JToolBar.HORIZONTAL);
      toolBar.add(add);
      toolBar.add(remove);
      toolBar.add(ok);
      toolBar.setFloatable(false);

      setLayout(new BorderLayout());
      add(barTop, BorderLayout.NORTH);
      add(new JScrollPane(table), BorderLayout.CENTER);
      add(toolBar, BorderLayout.SOUTH);
    }

    public void actionPerformed(ActionEvent e) {
      Object source = e.getSource();
      if(source == add) {
        ctm.addRow();
      }
      if(source == remove) {
        int[] rows = table.getSelectedRows();
        if(rows.length == 1) rows[0] = table.getSelectedRow();
        Arrays.sort(rows);//Very important!
        for(int i = rows.length - 1; i >= 0; i--) ctm.removeRow(rows[i]);//Must remove last first!
      }
      if(source == ok) {
        if(type.getText().equals("")) {//Not attribute is specified
          JOptionPane.showMessageDialog(
              this,
              Baobab.message.getString("ColorManagerNoTypeError"),
              Baobab.message.getString("error"),
              JOptionPane.ERROR_MESSAGE);
          return;
        }
        this.owner.setVisible(false);
      }
    }

    public class STDSTableModel extends AttributeValuesTableModel {

      public STDSTableModel() {
        data = new java.util.List[2];
        data[0] = new ArrayList();
        data[1] = new ArrayList();
        columnNames = new String[2];
        columnNames[0] = Baobab.message.getString("ColorManagerValues");
        columnNames[1] = Baobab.message.getString("ColorManagerColors");
      }
    }

    public Map colorTableModelToMap() {
      HashMap result = new HashMap();
      for(int i = 0; i < ctm.getRowCount(); i++) result.put(ctm.getValueAt(i, 0), ctm.getValueAt(i, 1));
      return result;
    }

    public ColoringMethod getColoringMethod() {
      ColoringMethod cm = null;
      try {
        cm = new ColoringTools.SingleTypeDescriptiveState(type.getText(), colorTableModelToMap());
      } catch(NumberFormatException ex) {
        JOptionPane.showMessageDialog(this, ex.getMessage(), "Error!", JOptionPane.ERROR_MESSAGE);
      }
      return cm;
    }

    protected JTextField type;
    protected STDSTableModel ctm;
    protected JTable table;
    protected JToolBar toolBar;
    protected JButton add, remove, ok;
  }




  public class STNVMethodPanel extends STDSMethodPanel {

    /**
     * STNV stands for:
     * Single Type with Numerical Value.
     */

    public STNVMethodPanel(Window owner) {
      super(owner);
      exact               = new JRadioButton(Baobab.message.getString("ColorManagerExact"));
      interval            = new JRadioButton(Baobab.message.getString("ColorManagerInterval"), true);
//      distribution        = new JRadioButton(message.getString("ColorManagerDistribution"));
//      reverseDistribution = new JRadioButton(message.getString("ColorManagerReverseDistribution"));
      ButtonGroup bg = new ButtonGroup();
                  bg.add(exact);
                  bg.add(interval);
//                  bg.add(distribution);
//                  bg.add(reverseDistribution);
                  JPanel panel = new JPanel(new FlowLayout(FlowLayout.LEFT));
                         panel.add(exact);
                         panel.add(interval);
//             panel.add(distribution);
//             panel.add(reverseDistribution);
                  toolBar.add(panel);

      auto = new JButton(Baobab.message.getString("ColorManagerAutoButton"));
      auto.setToolTipText(Baobab.message.getString("ColorManagerAutoButtonMnemo"));
      auto.addActionListener(new AutoActionListener());
      toolBar.add(auto);
      pack();
    }

    public class AutoActionListener implements ActionListener {
      public void actionPerformed(ActionEvent e) {
        class ParamDialog extends JDialog {
          public ParamDialog(Frame owner) {
            super(owner, true);
            setTitle(Baobab.message.getString("ColorManager.ParamDialogTitle"));
            beginLabel = new JLabel(Baobab.message.getString("ColorManager.ParamDialogBegin"));
            endLabel = new JLabel(Baobab.message.getString("ColorManager.ParamDialogEnd"));
            nbLabel = new JLabel(Baobab.message.getString("ColorManager.ParamDialogNb"));
            begin = new JTextField(6);
            end = new JTextField(6);
            nb = new JTextField(6);
            ok = new JButton(Baobab.message.getString("ok"));
            cancel = new JButton(Baobab.message.getString("cancel"));
            Container c = this.getContentPane();
            c.setLayout(new GridLayout(4, 2));
            c.add(beginLabel);
            c.add(begin);
            c.add(endLabel);
            c.add(end);
            c.add(nbLabel);
            c.add(nb);
            c.add(ok);
            c.add(cancel);
            ok.addActionListener(new ActionListener() {
              public void actionPerformed(ActionEvent ev) {
                setVisible(false);
              }
            });
            cancel.addActionListener(new ActionListener() {
              public void actionPerformed(ActionEvent ev) {
                nb.setText("0");
                setVisible(false);
              }
            });
            pack();
            GraphicsTools.center(this);
          }

          private JTextField begin, end, nb;
          private JLabel beginLabel, endLabel, nbLabel;
          private JButton ok, cancel;
        }
        ParamDialog d = new ParamDialog(parent);
        d.setVisible(true);
        try {
          double begin = Double.parseDouble(d.begin.getText());
          double end = Double.parseDouble(d.end.getText());
          int n = Integer.parseInt(d.nb.getText());
          if(n <= 0) return; //No interval or action cancelled.
          double step = (end - begin) / n;
          for(int i = 0; i <= n; i++) {
            float f = (float)i / (float)(n + 1);
            ctm.addRow("" + (begin + i*step), Color.getHSBColor(f, 1.0f, 1.0f));
          }
        } catch(NumberFormatException ex) {
          JOptionPane.showMessageDialog(
            owner,
            Baobab.message.getString("ColorManager.ParamDialogBadNumber"),
            Baobab.message.getString("error"),
            JOptionPane.ERROR_MESSAGE);
        }

      }

    }


    public ColoringMethod getColoringMethod() {
      int option = 0;
      if(exact.isSelected())               option = ColoringTools.SingleTypeNumericalValue.EXACT;
      if(interval.isSelected())            option = ColoringTools.SingleTypeNumericalValue.INTERVAL;
//      if(distribution.isSelected())        option = ColoringTools.SingleTypeNumericalValue.REPARTITION;
//      if(reverseDistribution.isSelected()) option = ColoringTools.SingleTypeNumericalValue.INVERSE_REPARTITION;
      return new ColoringTools.SingleTypeNumericalValue(type.getText(), colorTableModelToMap(), option);
    }

    protected JRadioButton exact, interval, distribution, reverseDistribution;
    protected JButton auto;

  }


  private Frame parent;
  private JTabbedPane methods;
}
