package bio.dendogram.GUI;

import gnu.jtools.stats.array.ArrayStatTools;
import gnu.jtools.utils.OperatingSystem;
import gnu.jtools.utils.gui.Range;
import bio.dendogram.*;
import bio.taxonomy.*;
import java.util.*;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Color;
import java.awt.Insets;
import java.awt.event.*;
import javax.swing.*;

public class TreeDrawing extends ScrollablePanel implements Cloneable, ActionListener {

  public TreeDrawing(Tree tree, Locale locale, OperatingSystem OS) {
    message = ResourceBundle.getBundle("bio/dendogram/GUI/bundle", locale);
    setLocale(locale);
    drawLengths         = false;
    drawBootstraps      = false;
    drawIds             = false;
    //drawGroups          = false;
    drawBranchesNumbers = false;
    drawClickableNodes  = false;
    drawLeavesNames     = true;
    drawingAlgorithm    = 0;
    clickableNodes = new ArrayList();
    climbUp = new JPopupMenu();
    climbDown = new JPopupMenu();
    climb = new JPopupMenu();
    climbUpSubMenu = new JMenu(message.getString("ClimbUp"));
    climbDownSubMenu = new JMenu(message.getString("ClimbDown"));
    climb.add(climbUpSubMenu);
    climb.add(climbDownSubMenu);
    setOS(OS);
    menuListener = new TreeDrawingActionListener(this);
    addMouseListener( new TreeDrawingMouseListener(this) );
    setMinimalLengthToDisplay(0.01);
    setMinimalLengthToClick(0.01);
    setFonts(DEFAULT_FONTS);
    setBackground(Color.white);
    setMargins(new Insets(this.DEFAULT_INSET_HEIGHT, this.DEFAULT_INSET_WIDTH,  this.DEFAULT_OFFSET_HEIGHT, this.DEFAULT_OFFSET_WIDTH));
    setTree(tree);
    setPreferredSize(100, 150); //Default values, prohibit NullPointerException when calling W or H.
    edited = false;
  }

  public Object clone() {
    TreeDrawing clone = new TreeDrawing(getTree(), getLocale(), getOS());
    clone.setActionCommand(this.getActionCommand());
    clone.setDrawBootstraps(this.drawBootstraps());
    clone.setDrawBranchesNumbers(this.drawBranchesNumbers());
    //clone.setDrawGroups(this.drawGroups());
    clone.setDrawLengths(this.drawLengths());
    clone.setDrawClickableNodes(this.drawClickableNodes());
    clone.setFonts(this.getFonts());
    clone.setMinimalLengthToClick(this.getMinimalLengthToClick());
    clone.setMinimalLengthToDisplay(this.getMinimalLengthToDisplay());
    clone.setPreferredSize(this.getPreferredSize());
    clone.setCurrentNode(this.getCurrentNode());
    clone.setDrawingAlgorithm(this.getDrawingAlgorithm());
    return clone;
  }

  public void addActionListener(ActionListener actionListener) {
    this.actionListener = actionListener;
  }
  public void setActionCommand(String actionCommand) {
    this.actionCommand = actionCommand;
  }
  public String getActionCommand() { return actionCommand; }

  public void setOS(OperatingSystem OS) { this.OS = OS; }

  public OperatingSystem getOS() { return OS; }

  /**
   * Actualizing the tree to be plotted:
   */

  public void actionPerformed(ActionEvent e) {
    //Called when a tree is modified.
    //check if the current node still exists:
    if(!tree.contains(currentNode)) returnToRoot();
    repaint();
    rebuildPopupMenus();
    scale();
    edited = true;
  }

  public void setMargins(Insets insets) {
    this.insetWidth = insets.left;
    this.offsetWidth = insets.right;
    this.insetHeight = insets.top;
    this.offsetHeight = insets.bottom;
  }

  public Insets getMargins() {
    return new Insets(insetHeight, insetWidth, offsetHeight, offsetWidth);
  }

  public void setCurrentNode(TreePath path) { setCurrentNode(tree.getNode(path)); }

  public void setCurrentNode(Node target) {
    currentNode = target;
    scale();
    repaint();
    rebuildPopupMenus();
    /**
     * if an ActionListener has been added, this throw an new ActionEvent object
     * in order to tell the listener that the current tree has changed.
     */
    try { actionListener.actionPerformed(new NewNodeEvent(this, 0, getActionCommand(), target)); }
    catch( NullPointerException ex) { /* no action listener added */ }
  }

  /**
   * Resize drawing according to numbe rof leaves and ranch lengths.
   * Must be called when branch lengths have been modified for instance.
   */
  public void scale() {
    fen = new Range(0, lengthMax(currentNode), 0, getNumberOfLeavesToBeDrawn(currentNode));
    fen.setWidth(L);  //Restore former H and L parameters.
    fen.setHeight(H);
    fen.setWHSet(insetWidth, offsetWidth, insetHeight, offsetHeight);
  }

  public void returnToRoot() { setCurrentNode(tree.getRootNode()); }

  public void climbUp(int whatBranch) { setCurrentNode(currentNode.getSon(whatBranch)); }
  public void climbDown(int numberOfLevels) {
    Node newNode = currentNode;
    for(int i = 0; i < -numberOfLevels; i++) newNode = newNode.getFather();
    setCurrentNode(newNode);
  }

  public Node getCurrentNode() { return currentNode; }

  public void setTree(Tree tree) {
    tree.addModificationListener(this);
    this.tree = tree;
    /**
     * Does not do setTree(new Vector(1, 1)), because setTree calls 'root' and is overdefined in derived classes.
     */
    currentNode = tree.getRootNode();
    scale();
    repaint();
    rebuildPopupMenus();
  }

  public Tree getTree() { return tree; }

  /**
   * Set the elements to be plotted.
   */
  public void setDrawBranchesNumbers (boolean yesOrNot) { drawBranchesNumbers = yesOrNot; }
  public void setDrawLengths         (boolean yesOrNot) { drawLengths         = yesOrNot; }
  public void setDrawBootstraps      (boolean yesOrNot) { drawBootstraps      = yesOrNot; }
  public void setDrawIds             (boolean yesOrNot) { drawIds             = yesOrNot; }
//  public void setDrawGroups          (boolean yesOrNot) { drawGroups          = yesOrNot; scale(); }
  public void setDrawClickableNodes  (boolean yesOrNot) { drawClickableNodes  = yesOrNot; }
  public void setDrawLeavesNames     (boolean yesOrNot) { drawLeavesNames     = yesOrNot; }
  public void setFont(int type, Font font) { fonts[type] = font; }
  public void setFonts(Font[] fonts) { this.fonts = fonts; }
  public void setDrawingAlgorithm(int i) { if(i >=0 && i <2) this.drawingAlgorithm = i; }
  public boolean drawBranchesNumbers () { return drawBranchesNumbers; }
  public boolean drawLengths         () { return drawLengths;         }
  public boolean drawBootstraps      () { return drawBootstraps;      }
  public boolean drawIds             () { return drawIds;             }
  //public boolean drawGroups          () { return drawGroups;          }
  public boolean drawClickableNodes  () { return drawClickableNodes;  }
  public boolean drawLeavesNames     () { return drawLeavesNames;     }
  public Font getFont(int type) { return fonts[type]; }
  public Font[] getFonts() { return fonts; }
  public int getDrawingAlgorithm() { return drawingAlgorithm; }
  public boolean isEdited() { return edited; }


  public void setMinimalLengthToDisplay(double lengthMin) { minimalLengthToDisplay = lengthMin; }
  public double getMinimalLengthToDisplay() { return minimalLengthToDisplay; }

  public void setMinimalLengthToClick(double lengthMin) { minimalLengthToClick = lengthMin; }
  public double getMinimalLengthToClick() { return minimalLengthToClick; }

  protected double length(Node node) {
    /**
     * The length to display, <> real length of the branch.
     */
    double lg = node.hasLength() ? node.getLengthValue() : 0;
    if( node != tree.getRootNode() && lg <= minimalLengthToDisplay ) lg = minimalLengthToDisplay;
    return lg; //The root node is the only one to have a 0 length.
  }

  protected double lengthMax(Node node) {
    /**
     * Maximal width of the whole subtree.
     */
    double result = 0;
    if(node.isLeaf()) {//leaf
      result = length(node);
    } else {//branch
      double[] x = new double[node.getNumberOfSons()];
      for(int i = 0; i < x.length; i++) {
        x[i] = lengthMax(node.getSon(i));
      }
      result = length(node) + ArrayStatTools.max(x);
    }
    return result;
  }

  protected double getVerticalCoord(Node node) {
//    **
//    * Quite the same than recursiveDrawTree, but draw nothing.
//    * It just send the vertical coordinate of the root of a particular subtree.
//    *
    double result = 0;
    double lg = length(node);
    if(node.isLeaf()) {
      result = TreeTools.getLeaves(currentNode).indexOf(node); //Send the leaf position in the list.
    } else {//branch
      double[] rooty = new double[node.getNumberOfSons()];
      Node subTree;
      for(int i = 0; i < rooty.length; i++) {
        subTree = node.getSon(i);
        //**
        //* Next step in recurtion.
        //*
        rooty[i] = getVerticalCoord(subTree);
      }
      result = ArrayStatTools.mean(rooty);
    }
    return result;
  }


  /**
   * NB: the three fallowing methods are overdefined from the JComponent class.
   */
  public void setPreferredSize(int Width, int Height) {
    /**
     * Set the scale.
     */
    setPreferredSize(new Dimension(Width, Height));
  }

  public void setPreferredSize(Dimension d) {
    /**
     * Set the scale.
     */
    this.L = d.width; this.H = d.height;
    fen.setWidth(L);
    fen.setHeight(H);
    fen.setWHSet(insetWidth, offsetWidth, insetHeight, offsetHeight);
    super.setPreferredSize(d);
    scale();
    repaint();
  }

  /**
   * ClickableNodes:
   */
  protected boolean isVisibilityGoodEnough(Node node) {
    return( length(node) >= minimalLengthToClick
            && tree.getNumberOfLeaves() / H >= 1/10 );
  }
//  protected boolean isVisibilityGoodEnough(Tree tree) {
//    return( currentTree.getNumberOfLeaves() < 15 );
//    }

  protected void addClickableNode(Node target, int x, int y) {
    /**
     * Add a new Double-Clickable node to the list:
     */
    ClickableNode cn = new ClickableNode(target);
    cn.defineArea(new Rectangle(x-CLICKABLE_AREA_SIZE/2,
                                y-CLICKABLE_AREA_SIZE/2,
                                CLICKABLE_AREA_SIZE,
                                CLICKABLE_AREA_SIZE
                                ));
    clickableNodes.add(cn);
  }

  protected void resetClickableNodes() { clickableNodes.clear(); }
  public ArrayList getClickableNodes() { return clickableNodes; }

  /**
   * Plotting methods:
   */
  protected void plotFirstBanchesNumbers2(Graphics g) {
    g.setFont(fonts[LEAVES]);
    double y = currentNode.getSon(0).getNumberOfLeaves() / 2;
    g.drawString("1", 3, fen.y(y) + 10);
    for(int i = 1; i < currentNode.getNumberOfSons(); i++) {//Do nothing if currentTree is a leaf.
      y = y + (currentNode.getSon(i - 1).getNumberOfLeaves() + currentNode.getSon(i).getNumberOfLeaves()) / 2;
      g.drawString("" + (i + 1), 3, fen.y(y) + 10);
    }
  }
  protected void plotFirstBanchesNumbers(Graphics g) {
    for(int i = 0; i < currentNode.getNumberOfSons(); i++) {//Do nothing if currentTree is a leaf.
      g.setFont(fonts[LEAVES]);
      g.drawString("" + (i + 1), 3, fen.y(getVerticalCoord(currentNode.getSon(i))) + 10);
    }
  }
  protected void plotLeafName(Graphics g, Node node, double x, double y) {
    if(!drawLeavesNames()) return;
    g.setFont(fonts[LEAVES]);
    g.drawString(node.getName(), fen.x(x) + 5, fen.y(y) + 3);
  }
  protected void plotGroupName(Graphics g, Node node, double x, double y) {
    if(!drawLeavesNames()) return;
    g.setFont(fonts[GROUPS]);
    //g.drawString(node.getName(), fen.x(x) + 5, fen.y(y) + 3);
    g.drawString(node.hasName() ? node.getName() : "" + node.getId(), fen.x(x) + 5, fen.y(y) + 3);
  }
  protected void plotLine(Graphics g, Node node, double x1, double y1, double x2, double y2) {
    g.drawLine(fen.x(x1), fen.y(y1), fen.x(x2), fen.y(y2));
  }
  protected void plotLengthValue(Graphics g, Node node, double x, double y) {
    if(!node.hasLength()) return;
    g.setFont(fonts[LENGTHS]);
    g.drawString( "" + node.getLength(), fen.x(x) + 3, fen.y(y) - 3);
  }
  protected void plotBootstrapValue(Graphics g, Node node, double x, double y) {
    if(!node.hasBootstrap()) return;
    g.setFont(fonts[BOOTSTRAPS]);
    g.drawString( "" + node.getBootstrap(), fen.x(x) + 3, fen.y(y) + 3);
  }
  protected void plotIds(Graphics g, Node node, double x, double y) {
    g.setFont(fonts[IDS]);
    String s = "#" + node.getId();
    int n = s.length();
    g.drawString(s, fen.x(x) - n*10 - 3, fen.y(y) - 5);
  }
  protected void plotClickableArea(Graphics g, double x, double y) {
    if(drawClickableNodes) g.drawRect( fen.x(x) - CLICKABLE_AREA_SIZE/2
                                     , fen.y(y) - CLICKABLE_AREA_SIZE/2
                                     , CLICKABLE_AREA_SIZE
                                     , CLICKABLE_AREA_SIZE
                                     );
  }

  protected List listOfLeavesToBeDrawn(Node node) {
    List result = new ArrayList();
    //if(node.isLeaf() || (GroupsManager.isGroup(node) && drawGroups)) result.add(node);
    if(node.isLeaf() || isGroup(node)) result.add(node);
    else for(int i = 0; i < node.getNumberOfSons(); i++) {
      result.addAll(listOfLeavesToBeDrawn(node.getSon(i)));
    }
    return result;
  }

  public static boolean isGroup(Node node) {
    boolean isCollapsable = Collapsable.class.isAssignableFrom(node.getClass());
    return isCollapsable && ((Collapsable)node).isCollapsed();
  }

  protected static int getNumberOfGroups(Node node) {
    int result = 0;
    if(isGroup(node) || node.isLeaf()) {//group
      result = 1;
    } else {//branch
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        result += getNumberOfGroups(node.getSon(i));
      }
    }
    return result;
  }

  protected int getNumberOfLeavesToBeDrawn(Node node) {
    //if(!drawGroups) return node.getNumberOfLeaves();
    //else return GroupsManager.getNumberOfGroups(node);
    return getNumberOfGroups(node);
  }

  protected void drawTree(Graphics g) {
    /**
     * Initiate plotting.
     */
    g.setFont(new Font("SansSerif", Font.PLAIN, 12));
    resetClickableNodes();
    recursiveDrawTree(g, 0, currentNode);
  }

  protected void drawTree2(Graphics g) {
    /**
     * Initiate plotting.
     */
    g.setFont(new Font("SansSerif", Font.PLAIN, 12));
    resetClickableNodes();
    double y = getNumberOfLeavesToBeDrawn(currentNode) / 2;
    recursiveDrawTree2(g, 0, y, currentNode);
  }

  protected void recursiveDrawTree2(Graphics g, double x, double y, Node node) {

    double xSon = x + length(node);
    plotLine(g, node, x, y, xSon, y);

    if(node.getNumberOfSons() > 0) {
      //if(GroupsManager.isGroup(node) && drawGroups) {
      if(isGroup(node)) {
        plotGroupName(g, node, xSon, y);
      } else {
        double[] ySon = new double[node.getNumberOfSons()];
        ySon[0] = y - (double)getNumberOfLeavesToBeDrawn(node) / 2 + (double)getNumberOfLeavesToBeDrawn(node.getSon(0)) / 2;
        recursiveDrawTree2(g, xSon, ySon[0], node.getSon(0));
        for(int i = 1; i < node.getNumberOfSons(); i++) {
          ySon[i] = ySon[i - 1] + ((double)getNumberOfLeavesToBeDrawn(node.getSon(i - 1)) + (double)getNumberOfLeavesToBeDrawn(node.getSon(i))) / 2;
          recursiveDrawTree2(g, xSon, ySon[i], node.getSon(i));
        }
        plotLine(g, node, xSon, ySon[0], xSon, ySon[ySon.length - 1]);
      }
    } else {
      plotLeafName(g, node, xSon, y);
    }

    /**
     * Plot lengths values:
     */
    try { if(drawLengths) plotLengthValue( g, node, x, y ); }
    catch(NullPointerException e) {}

    /**
     * Plot bootstraps values:
     */
    try { if(drawBootstraps) plotBootstrapValue( g, node, xSon, y ); }
    catch(NullPointerException e) {}

    /**
     * Plot ids:
     */
    if(drawIds) plotIds( g, node, xSon, y );

    /**
     * Double-Clickable nodes:
     */
    if(isVisibilityGoodEnough(node)) {
      addClickableNode( node, fen.x(x), fen.y(y) );
      plotClickableArea( g, x, y );
    }
  }

  protected double recursiveDrawTree(Graphics g, double rootx, Node node) {
    double result = 0;
    double lg = length(node);
    //if(node.isLeaf() || (drawGroups && GroupsManager.isGroup(node))) {
    if(node.isLeaf() || isGroup(node)) {
      result = listOfLeavesToBeDrawn(currentNode).indexOf(node); //Send the leaf position in the list.
      /**
       * Plot the names of the leaf:
       */
      if(node.isLeaf()) plotLeafName(g, node, rootx + lg, result);
      else plotGroupName(g, node, rootx + lg, result);
    } else {//branch
      double[] rooty = new double[node.getNumberOfSons()];
      Node subTree;
      for(int i = 0; i < rooty.length; i++) {
        subTree = node.getSon(i);
        /**
         * Next step in recursion.
         */
        rooty[i] = recursiveDrawTree(g, rootx + lg, subTree);
      }
      result = ArrayStatTools.mean(rooty);
      /**
       * Plot a vertical line.
       */
      plotLine( g, node, rootx+lg, rooty[0], rootx + lg, rooty[rooty.length-1] );
    }
    /**
     * Plot an horizontal line.
     */
    plotLine( g, node, rootx, result, rootx + lg, result );

    /**
     * Plot lengths values:
     */
    try { if(drawLengths) plotLengthValue( g, node, rootx, result ); }
    catch(NullPointerException e) {}

    /**
     * Plot bootstraps values:
     */
    try { if(drawBootstraps) plotBootstrapValue( g, node, rootx + lg, result ); }
    catch(NullPointerException e) {}

    /**
     * Plot ids:
     */
    if(drawIds) plotIds( g, node, rootx + lg, result );

    /**
     * Double-Clickable nodes:
     */
    if(isVisibilityGoodEnough(node)) {
      addClickableNode( node, fen.x(rootx), fen.y(result) );
      plotClickableArea( g, rootx, result );
    }
    return result;
  }

  /**
   * Draw the tree:
   */
  public void paintComponent(Graphics g) {
    super.paintComponent(g);
    if(drawingAlgorithm == 0) drawTree(g);
    if(drawingAlgorithm == 1) drawTree2(g);
    //if(drawBranchesNumbers && !(currentNode.isLeaf() || (currentNode.hasName() && drawGroups))) {
    if(drawBranchesNumbers && !(currentNode.isLeaf() || isGroup(currentNode))) {
      if(drawingAlgorithm == 0) plotFirstBanchesNumbers(g);
      if(drawingAlgorithm == 1) plotFirstBanchesNumbers2(g);
    }
  }

  public void paintPartOfComponent(Graphics g, Rectangle area) {
    // NB: area is the part of the drawing to be plotted, IN PIXELS(!!!)
    Range backUp = fen; //Actual range saved.
    fen = new Range(fen.xinv(area.x), fen.xinv(area.x + area.width), fen.yinv(area.y), fen.yinv(area.y + area.height));
    fen.setWidth(L);
    fen.setHeight(H);
    fen.setWHSet(insetWidth, offsetWidth, insetHeight, offsetHeight);
    g.setClip(0, 0, L, H);
    paintComponent(g);
    fen = backUp; //Old range restored.
  }

  /**
   * PopUp menus managing:
   */
  protected boolean canGoUp() {   //Criterion for moving toward leaves.
    //return !(currentNode.isLeaf() || (drawGroups && currentNode.hasName()));
    return !(currentNode.isLeaf() || isGroup(currentNode));
  }
  protected boolean canGoDown() { //Criterion for moving toward root.
    return currentNode != tree.getRootNode();
  }

  public void showClimbUpPopupMenu(int x, int y) {
    climbUp.show(this, x, y);
  }
  public void showClimbDownPopupMenu(int x, int y) {
    climbDown.show(this, x, y);
  }
  public void showClimbPopupMenu(int x, int y) {
    climb.show(this, x, y);
  }

  protected void rebuildPopupMenus() {
    climbUp.removeAll();
    climbUpSubMenu.removeAll();
    climbDown.removeAll();
    climbDownSubMenu.removeAll();

    String[] contenu;
    contenu = new String[currentNode.getNumberOfSons()];
    if(canGoUp()) {
      for(int i = 0; i < contenu.length; i++) {
        contenu[i] = message.getString("Branch") + " " + (i + 1);
      }
      for(int i = 0; i < contenu.length; i++) {
        JMenuItem it = new JMenuItem(contenu[i]);
        it.addActionListener(menuListener);
        it.setActionCommand("" + (i + 1));
        if(getOS().isMacOS()) climbUpSubMenu.add(it);
        else climbUp.add(it);
      }
    }
    if(canGoDown()) {//Can't move further...'
      contenu = new String[tree.getTreePath(currentNode).size()+1];
      contenu[0] = message.getString("ReturnToRoot");
      for(int i = 1; i < contenu.length; i++) {
        contenu[i] = ("" + (-i));
      }
      for(int i = 0; i < contenu.length; i++) {
        JMenuItem it = new JMenuItem(contenu[i]);
        it.addActionListener(menuListener);
        it.setActionCommand("" + (-i));
        if(getOS().isMacOS()) climbDownSubMenu.add(it);
        else climbDown.add(it);
      }
    }
  }

class TreeDrawingActionListener implements ActionListener {

  public TreeDrawingActionListener(TreeDrawing treeDrawing) {
    this.treeDrawing = treeDrawing;
  }

  public void actionPerformed(ActionEvent e) {
    String source = e.getActionCommand();
    try {
      int n = Integer.parseInt(source);
           if(n == 0) { returnToRoot(); } //Return to root
      else if(n > 0) { climbUp(n-1); }
      else { climbDown(n); }
    }
    catch(NumberFormatException ex) { System.out.println("Uncaught exception :-("); ex.printStackTrace(); }
  }

  private TreeDrawing treeDrawing;
}

class TreeDrawingMouseListener extends MouseAdapter {

  public TreeDrawingMouseListener(TreeDrawing treeDrawing) {
    this.treeDrawing = treeDrawing;
  }

  private void deepUncollapse(Node node) {
    if(node.isLeaf()) return;
    ((Collapsable)node).setCollapsed(false);
    for(int i = 0; i < node.getNumberOfSons(); i++) deepUncollapse(node.getSon(i));
  }

  public void mouseClicked(MouseEvent e) {
    Point p = e.getPoint();
    //Clicking on a node?
    List cna = getClickableNodes();
    //System.out.println("Click on: x=" + p.getX() + ", y=" + p.getY());
    for(int i = 0; i < cna.size(); i++) {
      ClickableNode cn = (ClickableNode)cna.get(i);
      if(cn.isInTargetArea(p)) {
        boolean isCollapsable = Collapsable.class.isAssignableFrom(cn.getTarget().getClass());
        if(e.isControlDown() & e.isShiftDown() && isCollapsable) {
          Collapsable col = (Collapsable) cn.getTarget();
          deepUncollapse(cn.getTarget());
          tree.treeIsModified("Subtree uncollapsed");
          return;
        }
        if(e.isControlDown() && isCollapsable) {
          Collapsable col = (Collapsable) cn.getTarget();
          col.setCollapsed(!col.isCollapsed());
          tree.treeIsModified("Node collapsed");
          return;
        }
        setCurrentNode(cn.getTarget());
        repaint();
        return;
      }
    }

    //Show navigation popup menus.
    if(OS.isMacOS()) {
      showClimbPopupMenu(e.getX(), e.getY());
    } else {
      if((e.getModifiers() & InputEvent.BUTTON3_MASK) != 0 && canGoUp()) showClimbUpPopupMenu(e.getX(), e.getY());
      if((e.getModifiers() & InputEvent.BUTTON1_MASK) != 0 && canGoDown()) showClimbDownPopupMenu(e.getX(), e.getY());
    }
  }

  private TreeDrawing treeDrawing;
}

  private ResourceBundle message;
  protected OperatingSystem OS;
  protected Node currentNode;
  protected Tree tree;
  private ArrayList clickableNodes;
  private JPopupMenu climbUp, climbDown, climb;
  private JMenu climbUpSubMenu, climbDownSubMenu;
  private TreeDrawingActionListener menuListener;
  protected boolean drawBranchesNumbers,
                    drawLengths,
                    drawBootstraps,
                    drawIds,
                    //drawGroups,
                    drawClickableNodes, //These state parameters are accessible in all derived classes.
                    drawLeavesNames;
  protected int drawingAlgorithm;

  protected double minimalLengthToDisplay, minimalLengthToClick;
  protected Range fen;
  protected int H, L;
  protected Font[] fonts;
  protected boolean edited;
  private ActionListener actionListener;
  private String actionCommand = "Current tree has changed"; //Default value
  private int insetWidth, offsetWidth, insetHeight, offsetHeight;

  public static final int CLICKABLE_AREA_SIZE = 7; //Pixels
  public static final int DEFAULT_INSET_WIDTH = 50;
  public static final int DEFAULT_INSET_HEIGHT = 15;
  public static final int DEFAULT_OFFSET_WIDTH = 200;
  public static final int DEFAULT_OFFSET_HEIGHT = 15;
  public static final Font[] DEFAULT_FONTS = { new Font("SansSerif", Font.PLAIN, 12)
                                             , new Font("SansSerif", Font.BOLD, 14)
                                             , new Font("SansSerif", Font.ITALIC, 10)
                                             , new Font("SansSerif", Font.ITALIC, 10)
                                             , new Font("SansSerif", Font.BOLD, 14)
                                             };
  public static final int LEAVES = 0;
  public static final int GROUPS = 1;
  public static final int LENGTHS = 2;
  public static final int BOOTSTRAPS = 3;
  public static final int IDS = 4;
}
