// -------------------------------------------------------------------------
//
// Copyright or © or Copr. CNRS, (November 16, 2004)
//
// This software is a library providing tool for bioinformatics.
//
// This software is governed by the CeCILL  license under French law and
// abiding by the rules of distribution of free software.  You can  use,
// modify and/ or redistribute the software under the terms of the CeCILL
// license as circulated by CEA, CNRS and INRIA at the following URL
// "http://www.cecill.info".
//
// As a counterpart to the access to the source code and  rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty  and the software's author,  the holder of the
// economic rights,  and the successive licensors  have only  limited
// liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading,  using,  modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean  that it is complicated to manipulate,  and  that  also
// therefore means  that it is reserved for developers  and  experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and,  more generally, to use and operate it in the
// same conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.
//
// -------------------------------------------------------------------------

package bio.dendogram;

/**
 * <p>Title: Dendogram</p>
 * <p>Description: Set of Tools for dealing with phylogenetic trees (dendograms).</p>
 * <p>Copyright: 2002</p>
 * @author Julien Dutheil
 * @version 1.0
 */

import gnu.jtools.stats.array.ArrayOperators;
import gnu.jtools.stats.list.ListOperators;
import gnu.jtools.utils.file.FileTools;
import gnu.jtools.utils.string.TextTools;

import java.util.*;
import java.io.*;

public class TreeTools {

  /**
   * Unroot a rooted tree.
   *
   * @param tree The tree to unroot.
   * @return Whether the tree could be unrooted or not.
   */
  public static boolean unroot(Tree tree) {
    Node root = tree.getRootNode();
    if(root.getNumberOfSons() == 2) {
      Node son1 = root.getSon(0);
      Node son2 = root.getSon(1);
      if(son1.isLeaf() && son2.isLeaf()) return false; // We can't unroot a single branch!
      // We manage to have a subtree in position 0:
      if(son1.isLeaf()) {
        root.swap(0, 1);
        son1 = root.getSon(0);
        son2 = root.getSon(1);
      }

      // Take care of branch lengths:
      if(son1.hasLength()) {
        if (son2.hasLength()) {
          // Both nodes have lengths, we sum them:
          son2.setLength(son1.getLengthValue() + son2.getLengthValue());
        } else {
          // Only node 1 has length, we set it to node 2:
          son2.setLength(son1.getLengthValue());
        }
        son1.deleteLength();
      } // Else node 2 may or may not have a branch length, we do not care!

      // Remove the root:
      root.removeSons();
      son1.addSon(son2);
      tree.setRootNode(son1);
      tree.actualizeNumberOfLeaves();
      tree.actualizeDimension();
      return true;
    } else return false; // Tree is already rooted.
  }

  public static Node getMonophyleticSubtreeThatContainsTheseSpecies(Node node, TaxonomicGroup tg) {
    Node result = null;
    if(listOfLeavesNames(node).containsAll(tg) && listOfLeavesNames(node).size() == tg.size()) result = node;
    else {
      for(int i = 0; i < node.getNumberOfSons() && result == null; i++) result = getMonophyleticSubtreeThatContainsTheseSpecies(node.getSon(i), tg);
    }
    return result;
  }

  public static Node getParaphyleticSubtreeThatContainsTheseSpecies(Node node, TaxonomicGroup tg) {
    Node result = null;
    if(listOfLeavesNames(node).containsAll(tg)) result = node;//One matching tree is found.
    for(int i = 0; i < node.getNumberOfSons() && result == null; i++) result = getParaphyleticSubtreeThatContainsTheseSpecies(node.getSon(i), tg);
      //If another smaller matching tree is found, result is now this smaller subtree.
    return result;
  }

  public static void insertASubTreeToEachTreeInAList(java.util.List treeList, Node subTree, TreePath completePath) {
    for(int i = 0; i < treeList.size(); i++) {
      ((Tree)treeList.get(i)).insertNode((Node)subTree.clone(), completePath);
    }
  }

  public static void rerootEachTreeInAList(java.util.List treeList, Node subTree) throws NodeException {
    for(int i = 0; i < treeList.size(); i++) {
      ((Tree)treeList.get(i)).newOutGroup(subTree);
    }
  }

  public static void unrootEachTreeInAList(java.util.List treeList) {
    for(int i = 0; i < treeList.size(); i++) {
      unroot((Tree)treeList.get(i));
    }
 }

 public static void unresolve(Node node) {
   for(int i = node.getNumberOfSons()-1; i >= 0; i--) {
     Node son = node.getSon(i);
     if(!son.isLeaf()) {
       node.removeSon(son);
       for (int j = 0; j < son.getNumberOfSons(); j++) {
         node.addSon(i, son.getSon(j));
       }
       //'son' is lost from here.
     }
   }
 }

 public static void deepUnresolve(Node node) {
   NodeList leaves = TreeTools.getLeaves(node);
   node.removeSons();
   for(int i = 0; i < leaves.size(); i++) {
     node.addSon((Node)leaves.get(i));
   }
 }

  public static boolean isThisPathASonOf(java.util.List son, java.util.List father) {
    boolean result = true;
    if(son.size()<father.size()) result = false;
    for(int i=0; i<father.size(); i++) result = result && father.get(i).equals(son.get(i));
    return result;
  }

  public static void correctName(Node node, String forbidenChars) {
    if(node.hasName()) {
      StringBuffer name = new StringBuffer(node.getName());
      for (int i = name.length() - 1; i >= 0; i--)if (forbidenChars.indexOf(
          name.charAt(i)) > -1) name.deleteCharAt(i);
      node.setName(name.toString());
    }
  }

  public static void correctNames(Node node, String forbidenChars) {
    correctName(node, forbidenChars);
    for(int i = 0; i < node.getNumberOfSons(); i++) correctNames(node.getSon(i), forbidenChars);
  }

  public static void correctNames(Tree tree, String forbidenChars) {
    correctNames(tree.getRootNode(), forbidenChars);
  }

  public static boolean checkLeavesNames(Tree tree, String chars) {
    TaxonomicGroup tg = tree.listOfLeavesNames();
    for(int i = 0; i < tg.size(); i++) {
      String name = (String)tg.get(i);
      if(TextTools.indexOfFirstOf(name, chars) > -1) return false;
    }
    return true;
  }

  public static NodeList getLeaves(Node node) {
    //Send all leaves of the parent node:
    NodeList result = new NodeList(10);
    if(node.isLeaf()) {
      result.add(node);
    } else {
      for(int i = 0; i < node.getNumberOfSons(); i++) result.addAll(getLeaves(node.getSon(i)));
    }
    return result;
  }

  public static void truncLeavesNames(Node node, int size) throws StringIndexOutOfBoundsException{
    if(size <= 0) throw new StringIndexOutOfBoundsException();
    if(node.isLeaf()) {
      if (node.getName().length() > size)
        node.setName(node.getName().substring(0, size));
    } else {
      for(int i = 0; i < node.getNumberOfSons(); i++) truncLeavesNames(node.getSon(i), size);
    }
  }
  public static void truncLeavesNames(Tree tree, int size) throws StringIndexOutOfBoundsException{
    truncLeavesNames(tree.getRootNode(), size);
  }

  public static Map loadTranslationMapFromFile(File file, int drop, char sep) throws IOException {
    BufferedReader br = new BufferedReader(new FileReader(file));
    String line = "";
    for(int i = 0; i < drop && line != null; i++) line = br.readLine();
    Map map = new HashMap();
    line = FileTools.getNextLine(br);
    while(line != null) {
      StringTokenizer st = new StringTokenizer(line, "" + sep);
      if(st.countTokens() != 2) throw new IOException("Invalid number of rows in file " + file.getAbsolutePath());
      map.put(st.nextToken(), st.nextToken());
      line = FileTools.getNextLine(br);
    }
    return map;
  }

  public static void translateNames(Node node, Map translation) {
    String newName = (String)translation.get(node.getName());
    if(newName != null && !newName.equals(node.getName())) node.setName(newName);
    for(int i = 0; i < node.getNumberOfSons(); i++) translateNames(node.getSon(i), translation);
  }
  public static void translateNames(Tree tree, Map translation) {
    translateNames(tree.getRootNode(), translation);
  }

  public static void translateLeavesNames(Node node, Map translation) {
      if(node.isLeaf()) {
        String newName = (String) translation.get(node.getName());
        if (newName != null && !newName.equals(node.getName())) node.setName(newName);
      }
      for(int i = 0; i < node.getNumberOfSons(); i++) translateLeavesNames(node.getSon(i), translation);
    }
  public static void translateLeavesNames(Tree tree, Map translation) {
    translateLeavesNames(tree.getRootNode(), translation);
  }

  public static Map reverseTranslation(Map translation) {
    HashMap reverse = new HashMap();
    Iterator i = translation.keySet().iterator();
    while(i.hasNext()) {
      Object key = i.next();
      Object value = translation.get(key);
      reverse.put(value, key);
    }
    return reverse;
  }


  public static boolean treeHasRepeatedLeavesNames(Tree tree) {
    int r = ArrayOperators.repeatedELements((String[])(tree.listOfLeavesNames().toArray(new String[0]))).length;
    return (r != 0);
  }

  public static boolean subtreeContainsNodeName(Node node, String nodeName) {
    if(node.isLeaf()) return node.getName().equals(nodeName);
    else {
      if(node.hasName() && node.getName().equals(nodeName)) return true;
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        if(subtreeContainsNodeName(node.getSon(i), nodeName)) return true;
      }
      return false;
    }
  }

  public static boolean treeContainsNodeName(Tree tree, String nodeName) {
    return subtreeContainsNodeName(tree.getRootNode(), nodeName);
  }

  public static boolean hasAtLeastOneLength(Node node) {
    boolean result = true;
    if(node.isLeaf()) result = node.hasLength();
    else {
      boolean[] b = new boolean[node.getNumberOfSons() + 1];
      b[0] = node.hasLength();
      for(int i = 1; i < b.length; i++) {
        b[i] = hasAtLeastOneLength(node.getSon(i - 1));
      }
      result = ArrayOperators.union(b);
    }
    return result;
  }
  public static boolean hasAtLeastOneLength(Tree tree) {
    return hasAtLeastOneLength(tree.getRootNode());
  }

  public static boolean hasAtLeastOneLengthIgnoreRoot(Tree tree) {
    Node root = tree.getRootNode();
    boolean[] result = new boolean[root.getNumberOfSons()];
    for(int i = 0; i < result.length; i++) {
      result[i] = hasAtLeastOneLength(root.getSon(i - 1));
    }
    return ArrayOperators.union(result);
  }//Do we really need this method?

  public static boolean hasAtLeastOneBootstrap(Node node) {
    boolean result = true;
    if(node.isLeaf()) result = false;
    else {
      boolean[] b = new boolean[node.getNumberOfSons() + 1];
      b[0] = node.hasBootstrap();
      for(int i = 1; i < b.length; i++) {
        b[i] = hasAtLeastOneBootstrap(node.getSon(i - 1));
      }
      result = ArrayOperators.union(b);
    }
    return result;
  }
  public static boolean hasAtLeastOneBootstrap(Tree tree) {
    return hasAtLeastOneBootstrap(tree.getRootNode());
  }

  public static boolean isMonophyletic(Node subtree, TaxonomicGroup tg) {
    boolean result = false;
    ArrayList leaves = TreeTools.listOfLeavesNames(subtree);
    if(leaves.containsAll(tg) && leaves.size() == tg.size()) result = true;
    else {
      for(int i = 0; i < subtree.getNumberOfSons() && result == false; i++) result = isMonophyletic(subtree.getSon(i), tg);
    }
    return result;
  }

  public static boolean isLeaf(Node node) { return node.getNumberOfSons() == 0; }

  public static boolean isRoot(Node node) { return node.getFather() == null; }

  public static boolean hasLengths(Node node) {
    boolean result = true;
    if(node.isLeaf()) result = node.hasLength();
    else {
      boolean[] b = new boolean[node.getNumberOfSons() + 1];
      b[0] = node.hasLength();
      for(int i = 1; i < b.length; i++) {
        b[i] = hasLengths(node.getSon(i - 1));
      }
      result = ArrayOperators.inter(b);
    }
    return result;
  }

  public static boolean hasBootstraps(Node node) {
    boolean result = true;
    if(node.isLeaf()) result = false;
    else {
      boolean[] b = new boolean[node.getNumberOfSons() + 1];
      b[0] = node.hasBootstrap();
      for(int i = 1; i < b.length; i++) {
        b[i] = hasBootstraps(node.getSon(i - 1)) || node.getSon(i - 1).isLeaf();
      }
      result = ArrayOperators.inter(b);
    }
    return result;
  }

  /****************************************************************************/

  public static int getNumberOfLeaves(Node node) {
    if(node.isLeaf()) return 1;
    else {
      int n = 0;
      for(int i = 0; i < node.getNumberOfSons(); i++) n += getNumberOfLeaves(node.getSon(i));
      return n;
    }
  }

  public static TaxonomicGroup listOfLeavesNames(Node node) {
    TaxonomicGroup result = new TaxonomicGroup(10);
    if(node.isLeaf()) {//a leaf
      if(node.hasName()) result.add(node.getName());
    } else {
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        result.addAll(listOfLeavesNames(node.getSon(i)));
      }
    }
    return result;
  }

  /****************************************************************************/

  public static int getDimension(Node node) {
    if(node.isLeaf()) return 0;
    else {
      int d = 0;
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        int sd = getDimension(node.getSon(i));
        if (sd > d)
          d = sd;
      }
      return d;
    }
  }

  /****************************************************************************/

  public static void actualizeNumberOfLeavesUp(Node node) {
    if(node.isLeaf()) node.setNumberOfLeaves(1);
    else {
      int n = 0;
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        actualizeNumberOfLeavesUp(node.getSon(i));
        n += node.getSon(i).getNumberOfLeaves();
      }
      node.setNumberOfLeaves(n);
    }
  }

  /****************************************************************************/

  public static void actualizeDimensionUp(Node node) {
    if(node.isLeaf()) node.setDimension(0);
    else {
      int d = 0;
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        actualizeDimensionUp(node.getSon(i));
        int sd = node.getDimension();
        if (sd > d) d = sd;
      }
      node.setDimension(d);
    }
  }

  /****************************************************************************/

  public static void actualizeNumberOfLeavesDown(Node node) {
    if(node.isLeaf()) node.setNumberOfLeaves(1);
    else {
      int n = 0;
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        n += node.getSon(i).getNumberOfLeaves();
      }
      node.setNumberOfLeaves(n);
    }
    if(node.hasFather()) actualizeNumberOfLeavesDown(node.getFather());
  }

  /****************************************************************************/

  public static void actualizeDimensionDown(Node node) {
    if(node.isLeaf()) node.setDimension(0);
    else {
      int d = 0;
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        int sd = node.getDimension();
        if (sd > d) d = sd;
      }
      node.setDimension(d);
    }
    if(node.hasFather()) actualizeDimensionDown(node.getFather());
  }

  /****************************************************************************/

  /**
   * Randomize the representation of a (sub)tree.
   *
   * This method does not modify the tree topology but randomly swap the order
   * of each son at a node.
   * This is done at the current node and for all the son node.
   *
   * @param node The node defining the subtree to randomize.
   */
  public static void randomizeDisplay(Node node) {
    // If this node is a leaf, we have nothing to do here...
    if(node.isLeaf()) return;
    // First we randomize each son node:
    for(int i = 0; i < node.getNumberOfSons(); i++) randomizeDisplay(node.getSon(i));
    // Then randomly permute branches:
    int n = node.getNumberOfSons();
    if(n == 1) {
      // Special case, nothing to do here!
      return;
    } else {
      NodeList sons = node.getSons();
      ListOperators.shuffle(sons);
      return;
    }
  }

  /**
   * This method call randomizeDisplay() on the root node.
   *
   * @param tree the tree to randomize.
   * @see randomizeDisplay(Node)
   */
  public static void randomizeDisplay(Tree tree) {
    randomizeDisplay(tree.getRootNode());
  }

  /****************************************************************************/

  private static boolean isLeaf(String chain) {
    return chain.indexOf(",")==-1 ? true : false;
  }

  private static String removeParenthesis(String chain) {
    String result = chain;
    int firstParenthesis = chain.indexOf("(") + 1;
    int lastParenthesis = chain.lastIndexOf(")");
    if (!isLeaf(chain))
      result = chain.substring(firstParenthesis, lastParenthesis);
    return result;
  }

  private static Vector extractElements(String chain) { //chain is a list of elements, separated by comas
    Vector elements = new Vector(2, 1);
    int comptPar = 0;
    int j = 0;
    for (int i = 0; i < chain.length(); i++) {
      if (chain.charAt(i) == '(')
        comptPar++;
      else if (chain.charAt(i) == ')')
        comptPar--;
      else if (comptPar == 0 && chain.charAt(i) == ',') { //premier �l�ment
        elements.addElement(chain.substring(j, i));
        j = i + 1;
      }
    }
    elements.addElement(chain.substring(j));
    return elements;
  }

  private static Node analyseAttributs(Tree tree, String chain) throws
      NumberFormatException {
    Node result = tree.createNode();
    if (!chain.equals("")) {
      StringTokenizer ana = new StringTokenizer(chain, ":");
      if (ana.countTokens() == 1) {
        if (chain.indexOf(":") > -1) { //only length
          result.setLength(Double.parseDouble(ana.nextToken()));
        }
        else { //only bootstrap
          result.setBootstrap(Double.parseDouble(ana.nextToken()));
        }
      }
      if (ana.countTokens() == 2) { //length and bootstrap
        result.setBootstrap(Double.parseDouble(ana.nextToken()));
        result.setLength(Double.parseDouble(ana.nextToken()));
      }
    }
    //neither length nor bootstrap, default values (cf. Tree class).
    return result;
  }

  public static Node parenthesisToNode(Tree tree, String chain) throws
      NumberFormatException { //chain is an element followed by its attributs.

    //analyse: is there some attributs(lengths, bootstraps)?

    int separateur;

    if (isLeaf(chain)) {
      separateur = chain.lastIndexOf(":");
      if (separateur == -1)
        separateur = chain.length();
    }
    else {
      separateur = chain.lastIndexOf(")") + 1;
    }

    String attribut, elements;

    if (separateur < chain.length()) {
      attribut = chain.substring(separateur);
      elements = removeParenthesis(chain.substring(0, separateur));
    }
    else { //no attributs
      attribut = "";
      elements = removeParenthesis(chain);
    }
//System.out.println("elements\t"+elements);
    Node v = analyseAttributs(tree, attribut);

    if (isLeaf(elements)) { //leaf
      v.setName(elements);
    }
    else { //branch
      Vector c = extractElements(elements);
      for (int i = 0; i < c.size(); i++) {
        tree.addNode(v, parenthesisToNode(tree, (String) (c.elementAt(i))));
      }
    }
    return v;
  }

  /****************************************************************************/

  private static String writeAttributs(Node node) {
    String result="";
    if(node.hasBootstrap()) result += node.getBootstrapValue();
    if(node.hasLength()) result += ":" + node.getLengthValue();
    return result;
  }

  public static String nodeToParenthesis(Node node) {
    String result="(";
    if(node.isLeaf()) {//leaf
      result = node.getName();
    } else {//branch
      Node subTree;
      for(int i = 0; i < node.getNumberOfSons() - 1; i++) {
        subTree = node.getSon(i);
        result += nodeToParenthesis(subTree)+", ";
      }
      subTree = node.getSon(node.getNumberOfSons() - 1);
      result += nodeToParenthesis(subTree) + ")";
    }
    return result + writeAttributs(node);
  }

  public static String treeToParenthesis(Tree tree) {
    return nodeToParenthesis(tree.getRootNode()) + ";";
  }

  /****************************************************************************/

  public static List getListOfSubNodes(Node node) {
    List subNodes = new ArrayList();
    subNodes.add(node);
    for(int i = 0; i < node.getNumberOfSons(); i++) subNodes.addAll(getListOfSubNodes(node.getSon(i)));
    return subNodes;
  }

  // @todo Implement this method... (Don't need this for now, but may prove useful...
  // pb: what about multifurcations?
  //public static Node getRandomTree(List leaveNames) {
  //
  //}

}
