// -------------------------------------------------------------------------
//
// Copyright or © or Copr. CNRS, (November 16, 2004)
//
// This software is a library providing tool for bioinformatics.
//
// This software is governed by the CeCILL  license under French law and
// abiding by the rules of distribution of free software.  You can  use,
// modify and/ or redistribute the software under the terms of the CeCILL
// license as circulated by CEA, CNRS and INRIA at the following URL
// "http://www.cecill.info".
//
// As a counterpart to the access to the source code and  rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty  and the software's author,  the holder of the
// economic rights,  and the successive licensors  have only  limited
// liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading,  using,  modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean  that it is complicated to manipulate,  and  that  also
// therefore means  that it is reserved for developers  and  experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and,  more generally, to use and operate it in the
// same conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.
//
// -------------------------------------------------------------------------

package bio.dendogram.xml;

/**
 * <p>Title: Dendogram</p>
 * <p>Description: Set of Tools for dealing with phylogenetic trees (dendograms).</p>
 * <p>Copyright: 2002</p>
 * <p>Company: </p>
 * @author not attributable
 * @version 1.0
 */

import bio.dendogram.*;
import bio.dendogram.xml.*;
import bio.dendogram.io.*;
import gnu.jtools.utils.attributes.Attribute;
import gnu.jtools.utils.storage.ReverseMap;
import java.util.*;
import java.io.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.w3c.dom.*;

public class XMLTreeFormat extends AbstractXMLOTree implements XMLITree {

  public static final String TREES_ROOT_NAME = "Trees";
  public static final String TREE_BLOCK      = "Tree block";
  public static final String TREE_NAME       = "Tree name";
  public static final String NODE            = "Node";

  public XMLTreeFormat(ReverseMap tagDescription) {
    this.tagDescription = tagDescription;
  }

  public XMLTreeFormat() {
    this.tagDescription = null;
  }

  public static class XTreeReader {

    public XTreeReader() throws ParserConfigurationException {
      factory = DocumentBuilderFactory.newInstance();
      builder = factory.newDocumentBuilder();
    }

    public XTree[] readXML(File inFile, ReverseMap translation) throws SAXException, IOException {
      this.translation = translation;
      ArrayList trees = new ArrayList();
      Document doc = builder.parse(inFile);
      Element rootElt = doc.getDocumentElement();
      String treeTagName = (String)translation.getValue(TREE_BLOCK);
      org.w3c.dom.NodeList treeElts = rootElt.getElementsByTagName(treeTagName);
      for(int i = 0; i < treeElts.getLength(); i++) {
        trees.add(treeElementToXTree((Element)treeElts.item(i)));
      }
      return (XTree[])trees.toArray(new XTree[0]);
    }

    private XTree treeElementToXTree(Element treeElement) {
      XTree tree = new XTree();
      org.w3c.dom.NodeList childNodeElements = treeElement.getChildNodes();

      List nodeElements = getChildrenElementsByTagName(treeElement, (String)translation.getValue(NODE));

      if(nodeElements.size() == 0) {
        //Tree is void
        System.err.println("WARNING: Void tree has been ignored.");
      } else if(nodeElements.size() == 1) {
        //This node is the root node:
        tree.setRootNode(nodeElementToXNode(tree, (Element)nodeElements.get(0)));
      } else {
        //The tree element correspond to the root node:
        bio.dendogram.Node root = tree.createNode();
        tree.setRootNode(root);
        for(int i = 0; i < nodeElements.size(); i++) {
          tree.addNode(root, nodeElementToXNode(tree, (Element)nodeElements.get(i)));
        }
      }
      //Look for tree attributes:
      Element treeNameElt = (Element)getChildrenElementsByTagName(treeElement, (String)translation.getValue(TREE_NAME)).get(0);
      if(treeNameElt != null) tree.setName(treeNameElt.getFirstChild().getNodeValue());
      tree.actualizeNumberOfLeaves();
      tree.actualizeDimension();
      return tree;
    }

    private XNode nodeElementToXNode(XTree tree, Element nodeElement) {
      XNode node = (XNode)tree.createNode();
      org.w3c.dom.NodeList nodeElements = nodeElement.getChildNodes();
      for(int i = 0; i < nodeElements.getLength(); i++) {
        org.w3c.dom.Node currentElement = nodeElements.item(i);
        switch(currentElement.getNodeType()) {
          case Element.ELEMENT_NODE: {
            if(((Element)currentElement).getTagName().equals(translation.getValue(NODE)))
              tree.addNode(node, nodeElementToXNode(tree, (Element)currentElement));
            else {
              String tagName = ((Element)currentElement).getTagName();
              if(tagName.equals(translation.getValue(XNode.NODE_NAME))) {
                // Set the name of the node:
                node.setName(getTextContent(currentElement), false);
              } else {
                Object key = translation.getKey(tagName);
                if (key != null) {
                  String content = getTextContent(currentElement);
                  //org.w3c.dom.Node attr = currentElement.getAttributes().getNamedItem("type");
                  //Object value;
                  //if (attr == null) { // Text
                  //  //try to parse the string:
                  //  System.out.print("'type' attribute not found in tag '" + tagName + "', trying to parse content...");
                  //  try {
                  //    value = new Double(content);
                  //    System.out.println(" type supposed to be 'Double'.");
                  //  } catch(NumberFormatException nfe1) {
                  //    try {
                  //      value = new Integer(content);
                  //      System.out.println(" type supposed to be 'Integer'.");
                  //    }
                  //    catch (NumberFormatException nfe2) {
                  //      try {
                  //        value = new Boolean(content);
                  //        System.out.println(" type supposed to be 'Boolean'.");
                  //      }
                  //      catch (NumberFormatException nfe3) {
                  //        value = content;
                  //        System.out.println(" failed. Default to string.");
                  //      }
                  //    }
                  //  }
                  //}
                  //else if (attr.getNodeValue().equals("integer")) {
                  //  value = new Integer(content);
                  //}
                  //else if (attr.getNodeValue().equals("double")) {
                  //  value = new Double(content);
                  //}
                  //else if (attr.getNodeValue().equals("boolean")) {
                  //  value = new Boolean(content);
                  //}
                  //else {
                  //  System.out.println("Unrecognized type '" + attr.getNodeValue() + "'.");
                  //  value = content;
                  //}
                  String value = new String(content);
                  if(key.equals(XNode.NODE_LENGTH)) node.getAttributes().put(key, new Double(value));
                  else if(key.equals(XNode.NODE_BOOTSTRAP)) node.getAttributes().put(key, new Double(value));
                  else node.getAttributes().put(key, value);
                }
              }
              //Otherwise node skipped.
            }
          }
          default : //Other nodes ignored.
        }
      }
      return node;
    }

    private DocumentBuilderFactory factory;
    private DocumentBuilder builder;
    private ReverseMap translation;
  }

  public static class XTreeWriter {

    public XTreeWriter() throws ParserConfigurationException, TransformerConfigurationException {
      factory = DocumentBuilderFactory.newInstance();
      builder = factory.newDocumentBuilder();
      transFactory = TransformerFactory.newInstance();
      transformer = transFactory.newTransformer();
    }

    public void writeXML(XTree tree, PrintWriter outFile, ReverseMap translation) throws TransformerException {
      this.translation = translation;
      document = builder.newDocument();
      Element root = document.createElement((String)translation.getValue(TREES_ROOT_NAME));
      document.appendChild(root);
      root.appendChild(xTreeToTreeElement(tree));
      transformer.transform(new DOMSource(document), new StreamResult(outFile));
    }

    public void writeXML(List trees, PrintWriter outFile, ReverseMap translation) throws TransformerException {
      this.translation = translation;
      document = builder.newDocument();
      Element root = document.createElement((String)translation.get(TREES_ROOT_NAME));
      document.appendChild(root);
      for(int i = 0; i < trees.size(); i++) root.appendChild(xTreeToTreeElement((XTree)trees.get(i)));
      transformer.transform(new DOMSource(document), new StreamResult(outFile));
    }

    public void writeXML(XTree[] trees, PrintWriter outFile, ReverseMap translation) throws TransformerException {
      this.translation = translation;
      document = builder.newDocument();
      Element root = document.createElement((String)translation.getValue(TREES_ROOT_NAME));
      document.appendChild(root);
      for(int i = 0; i < trees.length; i++) root.appendChild(xTreeToTreeElement(trees[i]));
      transformer.transform(new DOMSource(document), new StreamResult(outFile));
    }

    private Element xTreeToTreeElement(XTree tree) {
      this.translation = translation;
      Element elt = document.createElement((String)translation.getValue(TREE_BLOCK));
      if(tree.getName() != null) {
        //Give a name to this tree:
        Element nameElt = document.createElement((String)translation.getValue(TREE_NAME));
        nameElt.appendChild(document.createTextNode(tree.getName()));
        elt.appendChild(nameElt);
      }
      elt.appendChild(xNodeToNodeElement((XNode)tree.getRootNode()));
      return elt;
    }

    private Element xNodeToNodeElement(XNode node) {
      Element elt = document.createElement((String)translation.getValue(NODE));
      //Add son nodes:
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        elt.appendChild(xNodeToNodeElement((XNode)node.getSon(i)));
      }
      //Add attributes:
      Iterator i = node.getAttributes().entrySet().iterator();
      while(i.hasNext()) {
        Attribute attribute = (Attribute)i.next();
        String key = (String)translation.getValue(attribute.getKey());
        if(key != null) {
          Element attElt = document.createElement(key);
          Object value = attribute.getValue();
          //if(value.getClass() == Integer.class) {
          //  attElt.setAttribute("type", "integer");
          //} else if(value.getClass() == Double.class) {
          //  attElt.setAttribute("type", "double");
          //} else if(value.getClass() == Boolean.class) {
          //  attElt.setAttribute("type", "boolean");
          //}
          attElt.appendChild(document.createTextNode(value.toString()));
          elt.appendChild(attElt);
        }//Else attribute ignored.
      }
      return elt;
    }

    private DocumentBuilderFactory factory;
    private DocumentBuilder builder;
    private ReverseMap translation;
    private Document document;
    private TransformerFactory transFactory;
    private Transformer transformer;
  }

  /**
   * The content handler class:
   */
  private static class XMLReader extends DefaultHandler {

    private StringBuffer buffer;
    private Set tags;

    public XMLReader() {
      tags = new HashSet();
    }

    public void characters(char buf[], int offset, int len) {
      String s = new String(buf, offset, len);
      if (buffer == null)
        buffer = new StringBuffer(s);
      else
        buffer.append(s);
    }

    private String getText() {
      if(buffer == null) return "";
      String s = buffer.toString();
      buffer = null;
      return s;
    }

    public void startElement(String namespaceURI, String sName, String qName,
                             Attributes attrs) throws SAXException {
      String eName = sName;
      getText(); //empty buffer.
      if ("".equals(eName))
        eName = qName; //not namespaceaware
      tags.add(eName);
    }

    public void endElement(String namespaceURI, String sName, String qName) throws
        SAXException {
      // Nothing to do here
    }

    public Set getTags() { return tags; }
  }

  //****************************************************************************

   public static String getTextContent(org.w3c.dom.Node node) {
     String txt = "";
     if(node.getNodeType() == node.TEXT_NODE) txt += node.getNodeValue();
     else if(node.getNodeType() == node.ELEMENT_NODE) {
       org.w3c.dom.NodeList nl = ((Element)node).getChildNodes();
       for(int i = 0; i < nl.getLength(); i++) txt += getTextContent(nl.item(i));
     }
     return txt;
   }

   public static List getChildrenElementsByTagName(org.w3c.dom.Element element, String tagName) {
     ArrayList nodeElements = new ArrayList();
     org.w3c.dom.NodeList childNodeElements = element.getChildNodes();
     for(int i = 0; i < childNodeElements.getLength(); i++) {
       org.w3c.dom.Node currentElement = childNodeElements.item(i);
       switch(currentElement.getNodeType()) {
         case Element.ELEMENT_NODE: {
           if(((Element)currentElement).getTagName().equals(tagName)) nodeElements.add(currentElement);
         }
         default : //Other nodes ignored.
       }
     }
     return nodeElements;
   }

   public void writeTree(XTree tree, PrintWriter output) throws IOException {
     try {
       XTreeWriter xtw = new XTreeWriter();
       xtw.writeXML(tree, output, tagDescription);
     } catch(ParserConfigurationException pce) {
       throw new IOException(pce.getMessage());
     } catch(TransformerConfigurationException tce) {
       throw new IOException(tce.getMessage());
     } catch(TransformerException te) {
       throw new IOException(te.getMessage());
     }
   }

   public Tree[] readTree(File input) throws IOException {
     try {
       XTreeReader xtr = new XTreeReader();
       return xtr.readXML(input, tagDescription);
     } catch(ParserConfigurationException pce) {
       throw new IOException(pce.getMessage());
     } catch(SAXException se) {
       throw new IOException(se.getMessage());
     }
   }

   public static Set getAvailableTags(File file) throws IOException {
     SAXParserFactory factory = SAXParserFactory.newInstance();
     factory.setValidating(false);
     try {
       SAXParser parser = factory.newSAXParser();
       XMLReader handler = new XMLReader();
       parser.parse(file, handler);
       return handler.getTags();
     } catch(Exception e) {
       throw new IOException(e.getMessage());
     }
   }

}
