// -------------------------------------------------------------------------
//
// Copyright or © or Copr. CNRS, (November 16, 2004)
//
// This software is a library providing tool for bioinformatics.
//
// This software is governed by the CeCILL  license under French law and
// abiding by the rules of distribution of free software.  You can  use,
// modify and/ or redistribute the software under the terms of the CeCILL
// license as circulated by CEA, CNRS and INRIA at the following URL
// "http://www.cecill.info".
//
// As a counterpart to the access to the source code and  rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty  and the software's author,  the holder of the
// economic rights,  and the successive licensors  have only  limited
// liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading,  using,  modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean  that it is complicated to manipulate,  and  that  also
// therefore means  that it is reserved for developers  and  experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and,  more generally, to use and operate it in the
// same conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.
//
// -------------------------------------------------------------------------

package bio.dendogram.xml;

/**
 * <p>Title: Dendogram</p>
 * <p>Description: Set of Tools for dealing with phylogenetic trees (dendograms).</p>
 * <p>Copyright: 2002</p>
 * <p>Company: </p>
 * @author not attributable
 * @version 1.0
 */

import bio.dendogram.*;
import gnu.jtools.utils.attributes.Attribute;
import gnu.jtools.utils.attributes.AttributesMap;

public class XNode extends Node {

  protected AttributesMap attributes;

  public XNode(int id) {
    super(id);
    attributes = new AttributesMap();
  }

  public XNode(int id, String name) {
    super(id);
    attributes = new AttributesMap();
    setName(name);
  }

  public XNode(Node node) {
    super(node);
    attributes = new AttributesMap();
    if(node.hasName())      setName(node.getName());
    if(node.hasLength())    setLength(node.getLengthValue());
    if(node.hasBootstrap()) setBootstrap(node.getBootstrapValue());
    if(XNode.class.isAssignableFrom(node.getClass())) attributes = new AttributesMap(((XNode)node).attributes);
  }

  public Object clone() {
    XNode clone = new XNode((Node)super.clone());
    clone.attributes = new AttributesMap(attributes);
    return clone;
  }

  public void setName(String name) {
    //System.out.println("WARNING!!! This function [setName(String)] should not be used with XTree.");
    setName(name, false);
  }
  public void setName(String name, boolean isEditable) {
    super.setName(name);
    try {
      Attribute a = attributes.getAttribute(NODE_NAME);
                a.setValue(name);
                a.setNameEditable(false);
                a.setValueEditable(isEditable);
    } catch(NullPointerException e) {
      attributes.put(new Attribute(NODE_NAME, name, false, isEditable, String.class));
    }
  }

  public void setLength(double length) {
    //System.out.println("WARNING!!! This function [setLength(double)] should not be used with XTree.");
    setLength(length, true);
  }
  public void setLength(double length, boolean isEditable) {
    super.setLength(length);
    if(hasLength()) {
      Attribute a = attributes.getAttribute(NODE_LENGTH);
                a.setValue(new Double(length));
                a.setNameEditable(false);
                a.setValueEditable(isEditable);
    } else {
      attributes.put(new Attribute(NODE_LENGTH, new Double(length), false, isEditable, Double.class));
    }
  }

  public void setBootstrap(double bootstrap) {
    //System.out.println("WARNING!!! This function [setBootstrap(double)] should not be used with XTree.");
    setBootstrap(bootstrap, true);
  }
  public void setBootstrap(double bootstrap, boolean isEditable) {
    super.setBootstrap(bootstrap);
    if(hasBootstrap()) {
      Attribute a = attributes.getAttribute(NODE_BOOTSTRAP);
                a.setValue(new Double(bootstrap));
                a.setNameEditable(false);
                a.setValueEditable(isEditable);
    } else {
      attributes.put(new Attribute(NODE_BOOTSTRAP, new Double(bootstrap), false, isEditable, Double.class));
    }
  }

  public void deleteName() {
    super.deleteName();
    attributes.remove(NODE_NAME);
  }
  public void deleteLength() {
    super.deleteLength();
    attributes.remove(NODE_LENGTH);
  }
  public void deleteBootstrap() {
    super.deleteBootstrap();
    attributes.remove(NODE_BOOTSTRAP);
  }

  public String getName() {
    Attribute a = attributes.getAttribute(NODE_NAME);
    if(a != null) setName((String)a.getValue(), a.isValueEditable());
    else deleteName();
    return super.getName();
  }

  public Double getLength() {
//    try {
//      Attribute a = attributes.getAttribute(NODE_LENGTH);
//      double length = Double.parseDouble( (String) a.getValue());
//      setLength(length, a.isValueEditable());
//    } catch(Exception e) {
//      deleteLength();//No length!
//    }
//    return super.getLength();
    if(hasLength()) {
      try {
        return (Double)attributes.getAttribute(NODE_LENGTH).getValue();
      } catch(ClassCastException cce) {
        try {
          return new Double(attributes.getAttribute(NODE_LENGTH).getValue().toString());
        } catch(NumberFormatException nfe) {
          System.err.println("Warning: bad branch length at node " + getId() + ": " + attributes.getAttribute(NODE_LENGTH).getValue().toString());
          return null;
        }
      }
    } else return null;
  }

  public Double getBootstrap() {
//    try {
//      Attribute a = attributes.getAttribute(NODE_BOOTSTRAP);
//      double bootstrap = Double.parseDouble((String)a.getValue());
//      setBootstrap(bootstrap, a.isValueEditable());
//    } catch(Exception e) {
//      deleteBootstrap();//No bootstrap!
//    }
//    return super.getBootstrap();
    if(hasBootstrap()) {
      try {
        return (Double)attributes.getAttribute(NODE_BOOTSTRAP).getValue();
      } catch(ClassCastException cce) {
        try {
          return new Double(attributes.getAttribute(NODE_BOOTSTRAP).getValue().toString());
        } catch(NumberFormatException nfe) {
          System.err.println("Warning: bad bootstrap value at node " + getId() + ": " + attributes.getAttribute(NODE_BOOTSTRAP).getValue().toString());
          return null;
        }
      }
    } else return null;
  }

  public Attribute setAttribute(Attribute attribute) {
    return (Attribute)attributes.put(attribute);
  }

  public Attribute setAttribute(String name, Object value) {
    return (Attribute)attributes.put(name, value);
  }

  public void setAttributeNameEditable(String name, boolean isNameEditable) {
    try { attributes.getAttribute(name).setNameEditable(isNameEditable); }
    catch(NullPointerException e) {} //No attribute with this name, operation cancelled.
  }

  public void setAttributeValueEditable(String name, boolean isValueEditable) {
    try { attributes.getAttribute(name).setValueEditable(isValueEditable); }
    catch(NullPointerException e) {} //No attribute with this name, operation cancelled.
  }

  public Attribute setAttribute(String name, Object value, boolean isNameEditable, boolean isValueEditable) {
    return (Attribute)attributes.put(name, value, isNameEditable, isValueEditable);
  }

  public Object getAttribute(String name) {
    return attributes.get(name);
  }

  public AttributesMap getAttributes() {
    return attributes;
  }

  public void setAttributes(AttributesMap attributes) {
    this.attributes = attributes;
  }

  public int getNumberOfAttributes() {
    return attributes.size();
  }

  public Attribute removeAttribute(Attribute attribute) {
    return attributes.remove(attribute);
  }

  public Attribute removeAttribute(String name) {
    return attributes.remove(name);
  }

  /**
   * Tests:
   */
  public boolean hasName() {
    //if(XNode.class.isAssignableFrom(getClass())) return hasAttribute(XNode.NODE_NAME);
    //else return super.hasName();
    return hasAttribute(XNode.NODE_NAME);
  }
  public boolean hasLength() {
    //if(XNode.class.isAssignableFrom(getClass())) return hasAttribute(XNode.NODE_LENGTH);
    //else return super.hasLength();
    if(!hasAttribute(XNode.NODE_LENGTH)) return false;
    try {
      Double.parseDouble(attributes.getAttribute(NODE_LENGTH).getValue().toString());
    } catch(NumberFormatException nfe) {
      System.err.println("Warning: bad branch length at node " + getId() + ": " + attributes.getAttribute(NODE_LENGTH).getValue().toString());
      return false;
    }
    return true;
  }
  public boolean hasBootstrap() {
    //if(XNode.class.isAssignableFrom(getClass())) return hasAttribute(XNode.NODE_BOOTSTRAP);
    //else return super.hasBootstrap();
    if(!hasAttribute(XNode.NODE_BOOTSTRAP)) return false;
    try {
      Double.parseDouble(attributes.getAttribute(NODE_BOOTSTRAP).getValue().toString());
    } catch(NumberFormatException nfe) {
      System.err.println("Warning: bad bootstrap value at node " + getId() + ": " + attributes.getAttribute(NODE_BOOTSTRAP).getValue().toString());
      return false;
    }
    return true;
  }
  public boolean hasAttribute(String name) { return getAttribute(name) != null; }

  public static final String NODE_NAME      = "Node name";
  public static final String NODE_LENGTH    = "Branch length";
  public static final String NODE_BOOTSTRAP = "Bootstrap value";
}
