// -------------------------------------------------------------------------
//
// Copyright or © or Copr. CNRS, (November 16, 2004)
//
// This software is a library providing tool for bioinformatics.
//
// This software is governed by the CeCILL  license under French law and
// abiding by the rules of distribution of free software.  You can  use,
// modify and/ or redistribute the software under the terms of the CeCILL
// license as circulated by CEA, CNRS and INRIA at the following URL
// "http://www.cecill.info".
//
// As a counterpart to the access to the source code and  rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty  and the software's author,  the holder of the
// economic rights,  and the successive licensors  have only  limited
// liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading,  using,  modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean  that it is complicated to manipulate,  and  that  also
// therefore means  that it is reserved for developers  and  experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and,  more generally, to use and operate it in the
// same conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.
//
// -------------------------------------------------------------------------

package bio.dendogram.xml;

/**
 * <p>Title: Dendogram</p>
 * <p>Description: Set of Tools for dealing with phylogenetic trees (dendograms).</p>
 * <p>Copyright: 2002</p>
 * @author Julien Dutheil
 * @version 1.0
 */

import gnu.jtools.stats.table.Table;
import gnu.jtools.stats.table.TableRowNotFoundException;
import gnu.jtools.utils.attributes.Attribute;
import gnu.jtools.utils.attributes.AttributeNotFoundException;
import gnu.jtools.utils.attributes.AttributesMap;
import gnu.jtools.utils.attributes.BadAttributeNameException;
import gnu.jtools.utils.storage.CollectionTools;
import gnu.jtools.utils.storage.ReverseMap;
import bio.sequences.*;
import bio.sequences.xml.*;
import bio.dendogram.*;
import java.util.*;
import java.awt.Component;
import java.lang.reflect.*;
import javax.swing.*;

public class XTreeTools {

  /**
   * Tree with sequences
   */

  public static void loadSequences(XNode node, SequenceContainer sc) throws SequenceNotFoundException {
    // Leaves in the tree and sequences in the file must share the same name.
    if(node.isLeaf()) {
      XSequence seq = new XSequence(sc.getSequence(node.getName()));
      node.setAttribute(XSequence.SEQUENCE, seq, false, false);
      if(seq.getClass() == XSequence.class) {
        //Must add all attributes to the tree:
        node.getAttributes().putAll(((XSequence)seq).getAttributes());
      }
    } else {
      for(int i = 0; i < node.getNumberOfSons(); i++) loadSequences((XNode)node.getSon(i), sc);
    }
  }
  public static void loadSequences(XTree tree, SequenceContainer sc) throws SequenceNotFoundException {
    loadSequences((XNode)tree.getRootNode(), sc);
  }

  public static XSequence getSequence(XNode node) throws BadCharException {
    Object attribute = node.getAttribute(XSequence.SEQUENCE);
    if(attribute == null) return null;
    XSequence seq;
    try{ seq = (XSequence) attribute; }
    catch(ClassCastException cce) {
      // Sequence supposed to be a string!
      seq = new XSequence(node.getName(), attribute.toString(), new DefaultAlphabet());
    }
    // Dealing with attributes:
    AttributesMap attributes = new AttributesMap(node.getAttributes());
                  attributes.remove(XSequence.SEQUENCE);
    seq.getAttributes().putAll(attributes);
    return seq;
  }

  public static ArrayList getListOfSequences(XNode node) throws NullPointerException, BadCharException {
    ArrayList result = new ArrayList();
    XSequence seq = getSequence(node);
    if(seq != null) result.add(seq);
    for(int i = 0; i < node.getNumberOfSons(); i++) {
      result.addAll(getListOfSequences((XNode)node.getSon(i)));
    }
    return result;
  }

/*  public static void correctName(XNode node, String forbidenChars) {
    correctName(node, forbidenChars);
    SequenceTools.correctName(getSequence(node), forbidenChars);
  }

  public static void correctNames(XNode node, String forbidenChars) {
    correctName(node, forbidenChars);
    for(int i = 0; i < node.getNumberOfSons(); i++) correctName((XNode)node.getSon(i), forbidenChars);
  }*/

  public static SequenceContainer getSequenceContainer(XTree tree, Alphabet alphabet) throws Exception {
    return new VectorSequenceContainer(getListOfSequences((XNode)tree.getRootNode()), alphabet);
  }

  public static SequenceContainer getSequenceContainer(XNode subtree, Alphabet alphabet) throws Exception {
    return new VectorSequenceContainer(getListOfSequences(subtree), alphabet);
  }

  public static void loadAttributesFromTable(XNode node, Table table, ReverseMap tagDescription, boolean onlyLeaves) throws TableRowNotFoundException {
    if(onlyLeaves ? node.isLeaf() : node.hasName()) {
      List row = table.getRow(node.getName()); // May throw an exception here.
      for(int j = 0; j < row.size(); j++) {
        Object o = tagDescription.getKey(table.getColumnName(j));
        if( o != null) {
          Object value = row.get(j);
          if(value != null) node.attributes.put(o.toString(), value.toString(), true, true); //Otherwise attribute ignored.
        }
      }
    } else {
      for(int i = 0; i < node.getNumberOfSons(); i++)
        loadAttributesFromTable((XNode)node.getSon(i), table, tagDescription, onlyLeaves);
    }
  }

  public static void writeAttributesToTable(XNode node, Table table, ReverseMap tagDescription, boolean onlyLeaves) throws TableRowNotFoundException, AttributeNotFoundException {
    if(onlyLeaves ? node.isLeaf() : node.hasName()) {
      int row = table.getRowIndex(node.getName()); // May throw an exception here.
      String[] cols = table.getColumnNames();
      for(int j = 0; j < cols.length; j++) {
        Object o = tagDescription.getValue(cols[j]);
        if(o != null) {
          Object value = node.getAttribute(o.toString());
          if (value == null)throw new AttributeNotFoundException(cols[j]);
          table.setCell(row, j, value);
        } //Otherwise attribute ignored.
      }
    } else {
      for(int i = 0; i < node.getNumberOfSons(); i++)
        writeAttributesToTable((XNode)node.getSon(i), table, tagDescription, onlyLeaves);
    }
  }

  public static Set getAvailableTags(XNode node, boolean inner) {
    HashSet tags = new HashSet();
    // Addng tags from this node:
    ArrayList atts = new ArrayList(Arrays.asList(node.getAttributes().getNames()));
    LOOP: for(int i = atts.size() - 1; !inner && i >= 0; i--) {
      if(atts.get(i).toString().equals(XNode.NODE_NAME)) { atts.remove(i); continue LOOP; }
      if(atts.get(i).toString().equals(XNode.NODE_LENGTH)) { atts.remove(i); continue LOOP; }
      if(atts.get(i).toString().equals(XNode.NODE_BOOTSTRAP)) { atts.remove(i); continue LOOP; }
    }
    tags.addAll(atts);
    // Adding tags at son node:
    for(int i = 0; i < node.getNumberOfSons(); i++)
      tags.addAll(getAvailableTags((XNode)node.getSon(i), inner));
    return tags;
  }

  public static Set getAvailableTags(XTree tree, boolean inner) {
    return getAvailableTags((XNode)tree.getRootNode(), inner);
    // No tags are alowed in tree for now.
  }

  public static boolean nodeHasNonStandardAttributes(XNode node) {
    Set atts = getAvailableTags(node, false);
    if(!node.isLeaf()) atts.remove(XNode.NODE_NAME);
    atts.remove(XNode.NODE_LENGTH);
    atts.remove(XNode.NODE_BOOTSTRAP);
    return atts.size() > 0;
  }

  public static boolean subtreeHasNonStandardAttributes(XNode node) {
    boolean t = nodeHasNonStandardAttributes(node);
    for(int i = 0; i < node.getNumberOfSons(); i++) t = t || subtreeHasNonStandardAttributes((XNode)node.getSon(i));
    return t;
  }

  public static boolean treeHasNonStandardAttributes(XTree tree) {
    boolean t = subtreeHasNonStandardAttributes((XNode)tree.getRootNode());
    // No tags are alowed in tree for now.
    return t;
  }

  public static int getParsimonyScoreKnowingAncestralStates(XNode node, String characterName) throws AttributeNotFoundException {
    List nodes = TreeTools.getListOfSubNodes(node);
    nodes.remove(node);
    Iterator i = nodes.iterator();
    int nChange = 0;
    while(i.hasNext()) {
      XNode thisNode = (XNode)i.next();
      XNode father = (XNode)thisNode.getFather();
      Object state1 = father.getAttribute(characterName);
      if(state1 == null) throw new AttributeNotFoundException(characterName, "Character not found in node " + father.getName() + "(Id=" + father.getId() + ").");
      Object state2 = thisNode.getAttribute(characterName);
      if(state2 == null) throw new AttributeNotFoundException(characterName, "Character not found in node " + thisNode.getName() + "(Id=" + thisNode.getId() + ").");
      if(!state1.equals(state2)) nChange++;
    }
    return nChange;
  }

  public static int getParsimonyScoreKnowingAncestralStates(XTree tree, String characterName) throws AttributeNotFoundException {
    return getParsimonyScoreKnowingAncestralStates((XNode)tree.getRootNode(), characterName);
  }

  /**
   * First pass (top to bottom) of parsimony computation of ancestral states.
   *
   * This is a recursive method.
   * Ancestral states are stored as Collection object in a Map with each Node as a key.
   * The ancestral states for the current node are also sent as return of the function.
   *
   * @param node XNode The node at which we compute ancestral states.
   * @param characterName String The character for which we must compute ancestral states.
   * @param nodeStates Map Where to store ancestral states.
   * @throws NodeException If a node as less than two sons.
   * @throws AttributeNotFoundException If the character is not found for a leaf.
   * @return Collection All possible ancestral states for the current node.
   */
  public static Collection computeAncestralStatesUsingParsimonyFirstPass(XNode node, String characterName, Map nodeStates) throws NodeException, AttributeNotFoundException {
    Collection ancestralStates;
    if(node.isLeaf()) {
      ancestralStates = new HashSet();
      ancestralStates.add(node.getAttribute(characterName));
    } else {
      int n = node.getNumberOfSons();
      if(n < 2) throw new NodeException(node, "Node with less than 2 sons.");
      Collection[] ancestors = new Collection[n];
      ancestors[0] = computeAncestralStatesUsingParsimonyFirstPass((XNode)node.getSon(0), characterName, nodeStates);
      ancestors[1] = computeAncestralStatesUsingParsimonyFirstPass((XNode)node.getSon(1), characterName, nodeStates);
      ancestralStates = CollectionTools.getIntersection(ancestors[0], ancestors[1]);
      for(int i = 2; i < n; i++) {
        ancestors[i] = computeAncestralStatesUsingParsimonyFirstPass((XNode)node.getSon(i), characterName, nodeStates);
        ancestralStates = CollectionTools.getIntersection(ancestralStates, ancestors[i]);
      }
      if(ancestralStates.isEmpty()) {
        ancestralStates = CollectionTools.getUnion(ancestors[0], ancestors[1]);
        for(int i = 2; i < n; i++) {
          ancestralStates = CollectionTools.getUnion(ancestralStates, ancestors[i]);
        }
      }
    }
    nodeStates.put(node, ancestralStates);
    return ancestralStates;
  }

  public static void computeAncestralStatesUsingParsimonySecondPassInteractive(XNode node, String characterName, Map nodeStates, Component owner) throws AttributeNotFoundException {
    Collection ancestralStates = (Collection)nodeStates.get(node);
    String ancestralState;
    if(ancestralStates.size() == 1) {
      Iterator i = ancestralStates.iterator();
      ancestralState = new String(i.next().toString()); // Cloning object.
    } else {
      Object fatherChar = nodeStates.get(node.getFather());
      if(ancestralStates.contains(fatherChar)) {
        ancestralState = new String(fatherChar.toString());
      } else {
        // Ask for user choice:
        Object[] ancestors = ancestralStates.toArray();
        int choice = JOptionPane.showOptionDialog(owner,
            "Several ancestral states are possible in node " + node.getName() + "(id=" + node.getId() + "), pick one",
                                                  "Ancestral States",
                                                  JOptionPane.OK_OPTION,
                                                  JOptionPane.QUESTION_MESSAGE, null,
                                                  ancestors, ancestors[0]);
        ancestralState = new String(ancestors[choice].toString());
      }
    }
    node.setAttribute(new Attribute(characterName, ancestralState, true, true, String.class));
    // Recursion:
    for(int i = 0; i < node.getNumberOfSons(); i++)
      computeAncestralStatesUsingParsimonySecondPassInteractive((XNode)node.getSon(i), characterName, nodeStates, owner);
  }

  public static Collection computeNonAmbiguousAncestralStates(XNode node, String characterName) throws NodeException, AttributeNotFoundException {
    Collection ancestralStates;
    if(node.isLeaf()) {
      ancestralStates = new HashSet();
      ancestralStates.add(node.getAttribute(characterName));
    } else {
      int n = node.getNumberOfSons();
      if(n < 2) throw new NodeException(node, "Node with less than 2 sons.");
      Collection[] ancestors = new Collection[n];
      ancestors[0] = computeNonAmbiguousAncestralStates((XNode)node.getSon(0), characterName);
      ancestors[1] = computeNonAmbiguousAncestralStates((XNode)node.getSon(1), characterName);
      ancestralStates = CollectionTools.getUnion(ancestors[0], ancestors[1]);
      for(int i = 2; i < n; i++) {
        ancestors[i] = computeNonAmbiguousAncestralStates((XNode)node.getSon(i), characterName);
        ancestralStates = CollectionTools.getUnion(ancestralStates, ancestors[i]);
      }
      if(ancestralStates.size() == 1) {
        Object ancestralState = ancestralStates.iterator().next();
        node.setAttribute(characterName, ancestralState, true, true);
      } // Otherwise ambiguous!
    }
    return ancestralStates;
  }

  public static void setAttributeToAllSubtrees(XNode node, Attribute attribute) {
    node.getAttributes().put(attribute);
    for(int i = 0; i < node.getNumberOfSons(); i++) setAttributeToAllSubtrees(((XNode)node.getSon(i)), (Attribute)attribute.clone());
  }
  public static void setAttributeToAllSubtrees(XTree tree, Attribute attribute) {
    setAttributeToAllSubtrees((XNode)tree.getRootNode(), attribute);
  }

  public static void setAttributeToAllSubtrees(XNode node, String name, Object value) {
    node.getAttributes().put(name, value);
    for(int i = 0; i < node.getNumberOfSons(); i++) setAttributeToAllSubtrees(((XNode)node.getSon(i)),name, value);
  }
  public static void setAttributeToAllSubtrees(XTree tree, String name, Object value) {
    setAttributeToAllSubtrees((XNode)tree.getRootNode(), name, value);
  }

  public static void setAttributeNameEditableToAllSubtrees(XNode node, String name, boolean isNameEditable) {
    node.setAttributeNameEditable(name, isNameEditable);
    for(int i = 0; i < node.getNumberOfSons(); i++) {
      setAttributeNameEditableToAllSubtrees(((XNode)node.getSon(i)), name, isNameEditable);
    }
  }
  public static void setAttributeNameEditableToAllSubtrees(XTree tree, String name, boolean isNameEditable) {
    setAttributeNameEditableToAllSubtrees((XNode)tree.getRootNode(), name, isNameEditable);
  }

  public static void setAttributeValueEditableToAllSubtrees(XNode node, String name, boolean isValueEditable) {
    node.setAttributeValueEditable( name, isValueEditable);
    for(int i = 0; i < node.getNumberOfSons(); i++) {
      setAttributeValueEditableToAllSubtrees(((XNode)node.getSon(i)), name, isValueEditable);
    }
  }
  public static void setAttributeValueEditableToAllSubtrees(XTree tree, String name, boolean isValueEditable) {
    setAttributeValueEditableToAllSubtrees((XNode)tree.getRootNode(), name, isValueEditable);
  }

  public static void setAttributeToAllSubtrees(XNode node, String name, Object value, boolean isNameEditable, boolean isValueEditable) {
    node.getAttributes().put(name, value, isNameEditable, isValueEditable);
    for(int i = 0; i < node.getNumberOfSons(); i++) setAttributeToAllSubtrees((XNode)node.getSon(i), name, value, isNameEditable, isValueEditable);
  }
  public static void setAttributeToAllSubtrees(XTree tree, String name, Object value, boolean isNameEditable, boolean isValueEditable) {
    setAttributeToAllSubtrees((XNode)tree.getRootNode(), name, value, isNameEditable, isValueEditable);
  }

  public static boolean hasAttributeForEachSubtree(XNode node, String name) {
    boolean test = node.hasAttribute(name);
    for(int i = 0; i < node.getNumberOfSons(); i++) {
      test &= hasAttributeForEachSubtree((XNode)node.getSon(i), name);
    }
    return test;
  }

  public static boolean hasAttributeForEachSubtree(XTree tree, String name) {
    return hasAttributeForEachSubtree((XNode)tree.getRootNode(), name);
  }


  public static void removeAttributeFromAllSubtrees(XTree tree, String name) {
    removeAttributeFromAllSubtrees((XNode)tree.getRootNode(), name);
  }

  public static void removeAttributeFromAllSubtrees(XNode node, String name) {
    if(node.hasAttribute(name)) node.getAttributes().remove(name);
    for(int i = 0; i < node.getNumberOfSons(); i++) removeAttributeFromAllSubtrees((XNode)node.getSon(i), name);
  }

  public static void removeAttributeFromAllSubtrees(XTree tree, Attribute attribute) {
    removeAttributeFromAllSubtrees((XNode)tree.getRootNode(), attribute);
  }

  public static void removeAttributeFromAllSubtrees(XNode node, Attribute attribute) {
    if(node.hasAttribute(attribute.getName())) node.getAttributes().remove(attribute);
    for(int i = 0; i < node.getNumberOfSons(); i++) removeAttributeFromAllSubtrees((XNode)node.getSon(i), attribute);
  }

  /**
   * The value of each attribute must have a copy constructor defined otherwise an exception is thrown.
   *
   * @param node XNode
   * @param templateName String
   * @param copyName String
   * @throws NoSuchMethodException
   * @throws InstantiationException
   * @throws IllegalAccessException
   * @throws InvocationTargetException
   */
  public static void copyAttributeForAllSubtrees(XNode node, String templateName, String copyName)
  throws NoSuchMethodException, InstantiationException, IllegalAccessException, InvocationTargetException, BadAttributeNameException {
    if(node.hasAttribute(templateName)) {
      if(node.hasAttribute(copyName)) throw new BadAttributeNameException(copyName, "An attribute with this name already exists at node " + node.getId());
      Attribute att = (Attribute)node.attributes.getAttribute(templateName).deepClone();
      att.setName(copyName);
      att.setNameEditable(true);
      att.setValueEditable(true);
      node.setAttribute(att);
    }
    for(int i = 0; i < node.getNumberOfSons(); i++) copyAttributeForAllSubtrees((XNode)node.getSon(i), templateName, copyName);
  }

  public static void renameAttributeForAllSubtrees(XNode node, String oldName, String newName)
  throws BadAttributeNameException {
    if(node.hasAttribute(oldName)) {
      if(node.hasAttribute(newName)) throw new BadAttributeNameException(newName, "An attribute with this name already exists at node " + node.getId());
      else node.attributes.getAttribute(oldName).setName(newName);
    }
    for(int i = 0; i < node.getNumberOfSons(); i++) renameAttributeForAllSubtrees((XNode)node.getSon(i), oldName, newName);
  }

}
