package bio.sequences;

/**
 * <p>Title: Bio</p>
 * <p>Description: Tools for bioinformatics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Julien Dutheil
 * @version 1.0
 */

/**
 * @brief The Alphabet virtual class.
 * Interface for alphabets manipulation.
 *
 * An alphabet object defines all the states allowed for a particular type of
 * sequence. These states are coded as a charater (a string) and an integer.
 * The char description is the one found in the text descritpion of sequences,
 * typically in sequence files. This is the more intuitive. However, for
 * computionnal needs, this is often more efficient to store the sequences as
 * vector of integers. The link between the two descriptions is made <i>via</i>
 * the Alphabet classes, and the two methods intToChar() and charToInt().
 * The Alphabet interface also provides other tools, like getting the full name
 * of the states and so on.
 *
 * The alphabet objects may throw some exceptions derived of the AlphabetException
 * class.
 *
 * @see AlphabetException, BadCharException, BadIntException
 */
public interface Alphabet {

  /**
   * @brief Get the complete name of a state given its int description.
   *
   * In case of many bases with same number (i.e. N and X for nucleic alphabets),
   * this method will return the name of the first found in the vector.
   *
   * @param letter The int description of the given state.
   * @return The name of the state.
   */
  abstract String getName(int letter) throws BadIntException;

  /**
   * @brief Get the complete name of a state given its char description.
   *
   * In case of many bases with same number (i.e. N and X for nucleic alphabets),
   * this method will return the name of the first found in the vector.
   *
   * @param letter The char description of the given state.
   * @return The name of the state.
   */
  abstract String getName(String letter) throws BadCharException;

  /**
   * @name = Tests
   *
   * @{
   */

  /**
   * @brief Tell if a state (specified by its int descritption) is allowed by the
   * the alphabet.
   *
   * @param letter The int description.
   * @return 'True' if the state in known.
   */
  abstract boolean isIntInAlphabet(int letter);

  /**
   * @brief Tell if a state (specified by its char descritption) is allowed by the
   * the alphabet.
   *
   * @param letter The char description.
   * @return 'True' if the state in known.
   */
  abstract boolean isCharInAlphabet(String letter);

  /** @} */

  /**
   * @name Conversion methods
   *
   * @}
   */

  /**
   * @brief Give the char description of a state given its int description.
   *
   * @param letter The int description.
   * @return The char description.
   */
  abstract String intToChar(int letter) throws BadIntException;

  /**
   * @brief Give the int description of a state given its char description.
   *
   * @param letter The char description.
   * @return The int description.
   */
  abstract int charToInt(String letter) throws BadCharException;
  /** @} */

  /**
   * @name Sizes.
   *
   * @{
   */

  /**
   * @brief Get the number of allowed states in Alphabet (e.g. return 20 for DNA alphabet).
   * This is the number of characters used for state description.
   *
   * @return The total number of states.
   */
  abstract int getNumberOfChars();

  /**
   * @brief Get the number of <strong>distinct</strong> states in alphabet (e.g. return 15 for DNA alphabet).
   * This is the number of integers used for state description.
   *
   * @return The number of distinct states.
   */
  abstract int getNumberOfTypes();

  /**
   * @brief Get the number of <strong>resolved</strong> states in the alphabet (e.g. return 4 for DNA alphabet).
   * This is the method you'll need in most cases.
   *
   * @return The number of resolved states.
   */
  abstract int getSize();

  /** @} */

  // Some util:

  /**
   * @brief Identification method.
   *
   * Used to tell if two alphabet describe the same type of sequences.
   * For instance, this method is used by sequence containers to compare two alphabets and
   * autorize or deny addition of sequences.
   *
   * @return A text describing the alphabet.
   */
  abstract String getAlphabetType();
};
