package bio.sequences;

import bio.Vint;

/**
 * <p>Title: Bio</p>
 * <p>Description: Tools for bioinformatics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Julien Dutheil
 * @version 1.0
 */

import java.util.*;

/**
 * @brief The OrderedSequenceContainer interface.
 *
 * Interface to manage indexed containers.
 * Sequences may be accessed via their indice, i.e. their position in the container.
 */
public interface OrderedSequenceContainer extends SequenceContainer {

  /**
   * @brief Get the content of a sequence.
   *
   * @param i The position of the sequence.
   * @return The content of the sequence as a vector of integers.
   * @throw IndexOutOfBoundsException If the position does not match any sequence in the container.
   */
  public Vint getContent(int i) throws IndexOutOfBoundsException;

  /**
   * @brief Convert a particular sequence to a string.
   *
   * @param i The position of the sequence.
   * @return A string describing the content of the sequence.
   * @throw IndexOutOfBoundsException If the position does not match any sequence in the container.
   */
  public String toString(int i) throws IndexOutOfBoundsException;

  /**
   * @brief Retrieve a sequence object from the container.
   *
   * @param i The position of the sequence.
   * @return A pointer toward the Sequence with corresponding name.
   * @throw IndexOutOfBoundsException If the position does not match any sequence in the container.
   */
  public Sequence getSequence(int i) throws IndexOutOfBoundsException;

  /**
   * @brief Replace a sequence in the container.
   *
   * @param i         The position of the sequence.
   * @param sequence  The sequence to add.
   * @param checkName Tell if the container must check if the name of the sequence
   * is already used in the container before adding it.
   * @throw IndexOutOfBoundsException If the position does not match any sequence in
   * the container.
   * @throw Exception Any other kind of exception.
   */
  public void setSequence(int i, Sequence sequence, boolean checkNames) throws Exception;

  /**
   * @brief Extract (and remove) a sequence from the container.
   *
   * @param i The position of the sequence.
   * @throw IndexOutOfBoundsException If the name does not match any sequence in
   * the container.
   */
  public Sequence removeSequence(int i) throws IndexOutOfBoundsException;

  /**
   * @brief Delete a sequence of the container.
   *
   * @param i The position of the sequence.
   * @throw IndexOutOfBoundsException If the position does not match any sequence in
   * the container.
   */
  public void deleteSequence(int i) throws IndexOutOfBoundsException;

  /**
   * @brief Get the name of a particular sequence.
   *
   * @param i The position of the sequence.
   * @return The name of the sequence at position 'i'.
   * @throw IndexOutOfBoundsException If the position does not match any sequence in
   * the container.
   */
  public String getName(int i) throws IndexOutOfBoundsException;

  /**
   * @brief Get comments of a particular sequence.
   *
   * @param i The position of the sequence.
   * @return The comments associated to sequence at position 'i'.
   * @throw IndexOutOfBoundsException If the position does not match any sequence in
   * the container.
   */
  public List getComments(int i) throws IndexOutOfBoundsException;

  /**
   * @brief Set the comments of a particular sequence.
   *
   * @param i        The position of the sequence.
   * @param comments The comments to set to sequence with position 'i'.
   * @throw IndexOutOfBoundsException If the position does not match any sequence in
   * the container.
   */
  public void setComments(int i, List comments) throws IndexOutOfBoundsException;

  /**
   * @brief Get the position of a sequence in sequence container from its name.
   *
   * @param name The name of the sequence.
   * @return The position of the sequence with name 'name', if it exists.
   * @throw SequenceNotFoundException If no sequence with name 'name' could be found.
   */
  public int getSequencePosition(String name) throws SequenceNotFoundException;
}
