package bio.sequences;

/**
 * <p>Title: Bio</p>
 * <p>Description: Tools for bioinformatics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Julien Dutheil
 * @version 1.0
 */

import bio.*;

/**
 * @brief The Site class.
 *
 * Define specific attributes and methods for sites manipulation.
 * It is very similar to the Sequence object (a site is a vertical sequence!),
 * and characters at each position are coded as integers.
 * Sites have a 'position' attribute.
 * This attribute stands for an indice in a an alignment, and may be used as a unique identifier,
 * in the same manner that names identify sequence objects.
 * But for now, we do not allow to construct a Site directly from a string.
 * This should not be a constraint, since you never read sites directly from a file,
 * do you? You read sequences first...
 */
public class Site
    extends SymbolList {

  /**
   * @brief The position associated to this site.
   */
  protected int _position;

  /**
   * @brief Build a new void Site object with the specified alphabet.
   *
   * @param alpha The alphabet to use.
   */
  public Site(Alphabet alpha) {
    super(alpha);
    _position = 0;
  }

  /**
   * @brief Build a new void Site object with the specified alphabet and position.
   *
   * @param alpha    The alphabet to use.
   * @param position The position attribute for this site.
   */
  public Site(Alphabet alpha, int position) {
    super(alpha);
    _position = position;
  }

  /**
   * @brief Build a new Site object with the specified alphabet.
   * The content of the site is initialized from a vector of characters.
   *
   * @param site     The content of the site.
   * @param alpha    The alphabet to use.
   * @throw BadCharException If the content does not match the specified alphabet.
   */
  public Site(String[] site, Alphabet alpha) throws BadCharException {
    super(site, alpha);
    _position = 0;
  }

  /**
   * @brief Build a new Site object with the specified alphabet and position.
   * The content of the site is initialized from a vector of characters.
   *
   * @param site     The content of the site.
   * @param alpha    The alphabet to use.
   * @param position The position attribute for this site.
   * @throw BadCharException If the content does not match the specified alphabet.
   */
  public Site(String[] site, Alphabet alpha, int position) throws
      BadCharException {
    super(site, alpha);
    _position = position;
  }

  /**
   * @brief Build a new Site object with the specified alphabet.
   * The content of the site is initialized from a vector of integers.
   *
   * @param site     The content of the site.
   * @param alpha    The alphabet to use.
   * @throw BadIntException If the content does not match the specified alphabet.
   */
  public Site(Vint site, Alphabet alpha) throws BadIntException {
    super(site, alpha);
    _position = 0;
  }

  /**
   * @brief Build a new Site object with the specified alphabet and position.
   * The content of the site is initialized from a vector of integers.
   *
   * @param site     The content of the site.
   * @param alpha    The alphabet to use.
   * @param position The position attribute for this site.
   * @throw BadIntException If the content does not match the specified alphabet.
   */
  public Site(Vint site, Alphabet alpha, int position) throws BadIntException {
    super(site, alpha);
    _position = position;
  }

  /**
   * @brief The copy constructor.
   */
  public Site(Site site) {
    super(site);
    _position = site.getPosition();
  }

  /**
   * @name The Clonable interface
   *
   * @{
   */
  public Object clone() {
    return new Site(this);
  }

  /** @} */

  /**
   * @name Setting/getting the position of the site.
   *
   * @{
   */

  /**
   * @brief Get the position of this site.
   *
   * @return This site position.
   */
  public int getPosition() {
    return _position;
  }

  /**
   * @brief Set the position of this site.
   *
   * @param name The new position of the site.
   */
  public void setPosition(int position) {
    _position = position;
  }

}
