package bio.sequences;

/**
 * <p>Title: Bio</p>
 * <p>Description: Tools for bioinformatics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Julien Dutheil
 * @version 1.0
 */

import bio.*;

/**
 * @brief The SiteContainer interface.
 *
 * Container implementing the SiteContainer interface deal with <em>aligned</em> sequences.
 * This interface provides methods to retrieve, add or set sites in the alignment.
 * As for SequenceContainers, the maintenance of Sites is up to the container.
 * All site objects are cloned befored being added and retrieved.
 * All sites stored are deleted in the destructor of the container or after having called the deleteSite() method.
 */

public interface SiteContainer
    extends OrderedSequenceContainer {

  /**
   * @brief Get a site from the container.
   *
   * @param i The position of the site in the container.
   * @return A site objet corresponding to site i in the alignment.
   * @throw IndexOutOfBoundsException If the specified site does not exists.
   */
  public Site getSite(int i) throws IndexOutOfBoundsException;

  /**
   * @brief Set a site in the container.
   *
   * @param i             The position of the site in the container.
   * @param checkPosition Look if the position of the new site match a position attribute in the container.
   * @throw Exception If the specified site does not exists or is not correct.
   */
  public void setSite(int i, Site site, boolean checkPosition) throws Exception;

  /**
   * @brief Remove a site from the container.
   *
   * The site is not deleted, a pointer toward it is returned.
   *
   * @param i The position of the site in the container.
   * @return A pointer toward site i in the alignment.
   * @throw IndexOutOfBoundsException If the specified site does not exists.
   */
  public Site removeSite(int i) throws IndexOutOfBoundsException;

  /**
   * @brief Delete a site in the container.
   *
   * @param i The position of the site in the container.
   * @throw IndexOutOfBoundsException If the specified site does not exists.
   */
  public void deleteSite(int i) throws IndexOutOfBoundsException;

  /**
   * @brief Get the number of sites in the container.
   *
   * @return The number of sites in the container.
   */
  public int getNumberOfSites();

  /**
   * @brief Set all positions attributes.
   */
  public void reindexSites();

  /**
   * @brief Get all position attributes of sites.
   *
   * @return A vector with all site positions.
   */
  public Vint getSitePositions();
};
